# Copyright 2025 ghanvert. All rights reserved.
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#     http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.

from typing import Optional

import torch
from typing_extensions import Literal, override


_available_comparators = ["<", "<=", ">", ">=", "=="]


class Metric:
    def __init__(
        self, name: str, comparator: Literal["<", "<=", ">", ">=", "=="] = ">", main_metric: Optional[str] = None
    ):
        """
        Set a module to compute metrics.

        Args:
            name (`str`):
                Metric's module name.
            comparator (`str`, *optional*, defaults to `>`):
                Metric comparator to determine if current main metric value is the best calculated. Available
                options are: '<', '<=', '>', '>=' and '=='. For example, if set to '>', the comparation will be
                a > b, 'a' being current value and 'b' being previous value.
            main_metric (`str`, *optional*, defaults to `None`):
                Determine which is the main metric key in your compute output. By default, main metric key will be
                equal to the 'name' parameter.
        """
        self.name = name
        assert comparator in _available_comparators, f"Available options for comparator are: {_available_comparators}"
        self.comparator = comparator
        self.main_metric = main_metric if main_metric is not None else name

        # Lists of every argument, where every argument is also a list of tensors. Example:
        #   [[tensor, tensor, tensor], [tensor, tensor, tensor], ...]
        #   argument1                  argument2                 arguments...
        self.arguments = []

    @override
    def compute(self, *args: torch.Tensor) -> dict:
        """
        Compute metrics with the given arguments. This function returns a dictionary
        containing the main metric value and others.

        Example:
            ```
            def compute(self, predictions, references):
                # logic of how to calculate metrics here...

                return {
                    "accuracy": 0.85, # <-- this one is the main value
                    "f1": 0.89
                }
            ```

        NOTE: In the previous example, the main metric is 'accuracy', and its value is gonna be used along with
        'comparator' to compare if the metric is the best or not. By default, main metric is set to the name of
        the metric itself. You can change this behaviour with 'main_metric' on class initialization.
        """

    def _compute(self) -> dict:
        self._cat()
        output = self.compute(*self.arguments)
        self.clear()

        return output

    def clear(self):
        self.arguments.clear()

    def add_batch(self, *args: torch.Tensor):
        if len(self.arguments) == 0:
            # initialize lists
            self.arguments = [[] for _ in range(len(args))]

        for i, arg in enumerate(args):
            self.arguments[i].append(arg.cpu())  # transfer to CPU to avoid GPU memory issues

    def _cat(self):
        self.arguments = [torch.cat(arg) for arg in self.arguments]



# Copyright 2025 ghanvert. All rights reserved.
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#     http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.

from abc import ABC
from typing import Optional, Union

import torch
from accelerate import Accelerator
from typing_extensions import Any, override

from .states import TrainingState


class AcceleratorModule(ABC):
    """
    Super class to define training and validation logic without the need
    to write a training loop.

    The constructor of this class must implement `self.model`, specifying the model
    from `torch.nn.Module`. `self.teacher` is also a reserved property for teacher-student
    approaches.

    Methods:
        `forward` (*optional*):
            Defines the flow of data of model. If not implemented, `__call__`
            will not be possible (e.g. `self(...)`). Should return the model output.
        `training_step` (*optional*):
            Defines the training logic. Must return a loss `torch.Tensor` (scalar).
        `validation_step` (*optional*):
            Defines the validation logic. Must return a loss `torch.Tensor` (scalar).
            If not implemented, no validation will be executed.
        `collate_fn_train` (*optional*):
            Defines the collator function for train DataLoader.
        `collate_fn_val` (*optional*):
            Defines the collator function for validation DataLoader.
        `get_optimizer` (*optional*):
            Defines the optimizer. Must return the optimizer itself.
        `get_scheduler` (*optional*):
            Defines the scheduler. Must return the scheduler itself.
        `get_train_dataloader` (*optional*):
            Defines the train DataLoader. Must return a torch `DataLoader`.
        `get_validation_dataloader` (*optional*):
            Defines the validation DataLoader. Must return a torch `DataLoader`.

    Special methods (no implementation required):
        `__call__`:
            When calling this module, it will execute `forward` method.
        `__repr__`:
            When reproducing this module (e.g. Jupyter Notebook cell), this will print
            the model structure from `torch.nn.Module` specified in `self.model`.
        `__str__`:
            When printing this module or using it as a `str` type, this will represent
            the `torch.nn.Module` specified in `self.model`.
        `__len__`:
            When casting this module with `len` Python function, it will return the
            number of parameters of the base model specified in `self.model` from
            `torch.nn.Module`.
    """

    _implemented_collate_fn_train = False
    _implemented_collate_fn_val = False
    _accelerator: Accelerator = None
    _log_every: int = 1
    _extended = False
    state: TrainingState = None
    device: torch.device = None
    status_dict: dict = None
    batch_size: Union[int, tuple[int, int]] = None

    @override
    def forward(self, *args: Any, **kwargs: Any) -> torch.Tensor:
        """Defines the flow of data."""

    @override
    def training_step(self, batch: Any) -> torch.Tensor:
        """Defines the training logic. Must return a loss tensor (scalar)."""

    @override
    def validation_step(self, batch: Any) -> dict:
        """
        Defines the validation logic. Must return a dictionary containing
        each metric with predictions and targets, and also the loss value in the dictionary.

        Example:
            ```
            # format is ==> "metric": (predictions, targets)
            return {
                "loss": validation_loss_tensor, # (scalar tensor)
                # with additional metrics:
                "accuracy": (accuracy_predictions, accuracy_targets),
                "bleu": (bleu_predictions, bleu_targets)
            }
            ```
        """

    @override
    def collate_fn_train(self, batch: list) -> Any:
        """Defines a collate function for PyTorch train DataLoader."""

    @override
    def collate_fn_val(self, batch: list) -> Any:
        """Defines a collate function for PyTorch validation DataLoader."""

    @override
    def get_optimizer(self, *args: Any, **kwargs: Any) -> Any:
        """Defines a custom PyTorch optimizer logic here."""

    @override
    def get_scheduler(self, optimizer: Any, steps_per_epoch: int, epochs: int) -> Any:
        """Defines a custom PyTorch scheduler logic here."""

    @override
    def get_train_dataloader(self, *args: Any, **kwargs: Any) -> Any:
        """Defines a custom PyTorch DataLoader class for training."""

    @override
    def get_validation_dataloader(self, *args: Any, **kwargs: Any) -> Any:
        """Defines a custom PyTorch DataLoader class for validation."""

    def log(self, values: dict, log_kwargs: dict | None = {}):
        if self._accelerator.is_main_process:
            train_or_eval = "global_step" if self.model.training else "eval_global_step"
            if (self.status_dict[train_or_eval] + 1) % self._log_every == 0:
                self._accelerator.log(values, step=self.status_dict[train_or_eval], log_kwargs=log_kwargs)

    def __init_subclass__(cls, **kwargs):
        if (
            cls.training_step == AcceleratorModule.training_step
            and cls.validation_step == AcceleratorModule.validation_step
        ):
            raise TypeError("Subclasses of 'Trainer' must override 'training_step' and/or 'validation_step' methods.")

        if cls.collate_fn_train != AcceleratorModule.collate_fn_train:
            cls._implemented_collate_fn_train = True

        if cls.collate_fn_val != AcceleratorModule.collate_fn_val:
            cls._implemented_collate_fn_val = True

        super().__init_subclass__(**kwargs)

    def __call__(self, *args: Any, **kwargs: Any):
        return self.forward(*args, **kwargs)

    def __repr__(self):
        return self.model

    def __str__(self):
        return self.model.__repr__()

    def __len__(self):
        return sum(p.numel() for p in self.model.parameters())

    @classmethod
    def from_hf(cls, path: str, type: Union[str, Any] = None, **kwargs: Optional[Any]):
        """
        Build a custom AcceleratorModule for HuggingFace's transformers library. It simply replaces the following standard:

        ```
        class Module(AcceleratorModule):
            def __init__(self):
                self.model = AutoModel.from_pretrained(path, **kwargs)

            def training_step(self, batch):
                return self.model(**batch).loss

            def validation_step(self, batch):
                return {"loss": self.model(**batch).loss}
        ```

        Args:
            path (`str`):
                Path for HuggingFace model.
            type (`str` or `Any`):
                Model type in transformers library. It can be the class itself or a string (no need for imports).
            kwargs (`Any`):
                Keyword arguments for `from_pretrained` function for model initialization.
        """
        if isinstance(type, str):
            import importlib

            module = importlib.import_module("transformers")
            type = getattr(module, type)
        elif type is None:
            from transformers import AutoModel

            type = AutoModel

        class Module(AcceleratorModule):
            def __init__(self):
                self.model = type.from_pretrained(path, **kwargs)

            def training_step(self, batch):
                return self.model(**batch).loss

            def validation_step(self, batch):
                return self.model(**batch).loss

        return Module()


class ExtendedAcceleratorModule(AcceleratorModule):
    """
    Extended module from `AcceleratorModule` to enhance `training_step` function. This
    means that the backpropagation part must be done manually.

    Example:
        ```
        class Module(ExtendedAcceleratorModule):
            # other logic remains the same

            def training_step(self, batch):
                loss = ...
                self.backward(loss)
                self.step_optimizer()
                self.step_scheduler()

                return loss  # loss will only be used to log metrics.
        ```

    NOTE: `grad_accumulation_steps` in `fit` function from `Trainer` will not work. If you want to accumulate gradients
    and then backpropagate, you may want to make use of `self.status_dict["epoch_step"]`.
    """

    _extended = True

    def backward(self, loss: torch.Tensor, **kwargs):
        """
        Performs backward operation.

        Args:
            `loss` (`torch.Tensor`):
                Scalar loss tensor to backward.
            `kwargs` (`Any`):
                Extra arguments to be passed to 'accelerator.backward' function.
        """
        self._accelerator.backward(loss, **kwargs)

    def step_optimizer(self):
        self.state.optimizer.step()

    def step_scheduler(self):
        self.state.scheduler.step()

    def step(self):
        """Step optimizer and scheduler (in that order). If there is no scheduler, it will be ignored."""
        self.step_optimizer()
        if self.state.scheduler is not None:
            self.step_scheduler()

    def zero_grad(self, set_to_none: bool = True):
        """
        Call optimizer's 'zero_grad' operation to reset gradients.

        Args:
            `set_to_none` (`bool`, *optional*, defaults to `True`):
                Set gradients to `None` instead of `0`.
        """
        self.state.optimizer.zero_grad(set_to_none=set_to_none)

    @override
    def training_step(self, batch: Any):
        pass



# Copyright 2025 ghanvert. All rights reserved.
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#     http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.

import os
from typing import Union

import numpy as np
import pandas as pd
import torch
import torch.nn.functional as F
from typing_extensions import Any


WORLD_SIZE = int(os.getenv("WORLD_SIZE", 1))
RANK = int(os.getenv("RANK", 0))
MASTER_PROCESS = RANK == 0
LAST_PROCESS = RANK == WORLD_SIZE - 1


def _get_array_partition(array: Any, n_partitions: int, get: int) -> Any:
    """
    Get an specific partition of an array-like object. This returns a view of the
    original object.

    Args:
        array (`Any`):
            Array-like.
        n_partitions (`int`):
            Number of partitions to create from the original array-like.
        get (`int`):
            Get the specific partition.

    Returns:
        `Any`: Specific partition (`get`) of the given array-like.
    """
    length = len(array)
    chunk_size, remainder = divmod(length, n_partitions)

    start = get * chunk_size + min(get, remainder)
    end = start + chunk_size + (1 if get < remainder else 0)

    return array[start:end]


def prepare_dataframe(df: pd.DataFrame, even: bool = False) -> tuple[pd.DataFrame, int]:
    """
    Get an specific partition of a Pandas DataFrame per process.

    Args:
        df (`pd.DataFrame`):
            Pandas DataFrame.
        even (`bool`, *optional*, defaults to `False`):
            Wether to create even partitions across all processes.

    Returns:
        `list`: Rank-specific Pandas DataFrame partition.
    """
    remainder = 0
    if WORLD_SIZE > 1:
        partition_size, remainder = divmod(len(df), WORLD_SIZE)

        for rank, i in enumerate(range(0, len(df), partition_size + remainder)):
            if rank == RANK:
                df = df.iloc[i : i + partition_size + remainder]
                break

        if even and LAST_PROCESS and remainder != 0:
            last_row = df.iloc[-1:]
            df = pd.concat([df] + [last_row] * (remainder * (WORLD_SIZE - 1)), ignore_index=True)

        remainder = remainder * (WORLD_SIZE - 1)

    return df, remainder


def prepare_array(array: Any, even: bool = False) -> Any:
    """
    Get an specific partition of an array-like per process.

    Args:
        array (`Any`):
            Array-like.
        even (`bool`, *optional*, defaults to `False`):
            Wether to create even partitions across all processes.

    Returns:
        `Any`: Rank-specific array-like partition.
    """
    if not hasattr(array, "__len__"):
        raise ValueError("Values passed to 'prepare_array' must be array-like only (not scalar values).")

    if WORLD_SIZE > 1:
        array_type = type(array)
        if even:
            size, remainder = divmod(len(array), WORLD_SIZE)
            max_size = size + remainder
            diff = len(array) - max_size

            last_elem = array[-1:]
            padding_array = np.repeat(last_elem, diff, axis=0)
            array = np.concatenate([array, padding_array], axis=0)

        array = _get_array_partition(array, n_partitions=WORLD_SIZE, get=RANK)
        if isinstance(array, np.ndarray):
            if array_type is torch.Tensor:
                array = torch.from_numpy(array)
            elif array_type is list:
                array = array.tolist()

    return array


def prepare(*objs, even: bool = False) -> list:
    """
    Get an specific partition for every object per process.

    Args:
        objs (`Any`):
            Objects to prepare.
        even (`bool`, *optional*, defaults to `False`):
            Wether to create even partitions across all processes.

    Returns:
        `list`: List of rank-specific object partitions.
    """
    prepared = []
    for obj in objs:
        if isinstance(obj, pd.DataFrame):
            prepared.append(prepare_dataframe(obj, even=even))
        elif hasattr(obj, "__len__"):
            prepared.append(prepare_array(obj, even=even))
        else:
            prepared.append(obj)

    return prepared if len(prepared) > 1 else prepared[0]


def divide_into_batches(lst, batch_size):
    return [lst[i : i + batch_size] for i in range(0, len(lst), batch_size)]


def iterbatch(lst, batch_size):
    for i in range(0, len(lst), batch_size):
        yield lst[i : i + batch_size]


def pad(tensor: torch.Tensor, to: int, value: Union[float, int]):
    size = tensor.shape[-1]
    if size >= to:
        return tensor

    pad_size = to - size
    tensor = F.pad(tensor, (0, pad_size), value=value)

    return tensor



# Copyright 2025 ghanvert. All rights reserved.
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#     http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.

from typing import Any, Union

import numpy as np
import torch
from torch.utils.data._utils.collate import default_collate
from transformers.tokenization_utils_base import BatchEncoding


def collate_tokenizer_inputs(batch: list, pad_token_id: int, label_pad_token_id: int, padding_side: str):
    include_attention_mask = "attention_mask" in batch[0]
    include_labels = "labels" in batch[0]

    inputs = []
    labels = []
    for feature in batch:
        inputs.append(len(feature["input_ids"]))
        if include_labels:
            labels.append(len(feature["labels"]))

    max_input_length = max(inputs)
    if include_labels:
        max_label_length = max(labels)

    inputs = []
    attention_masks = []
    labels = []
    for feature in batch:
        inputs_remainder = [pad_token_id] * (max_input_length - len(feature["input_ids"]))
        if include_attention_mask:
            attention_masks_remainder = [0] * (max_input_length - len(feature["input_ids"]))
        if include_labels:
            labels_remainder = [label_pad_token_id] * (max_label_length - len(feature["labels"]))

        if include_labels and isinstance(feature["labels"], list):
            feature = {
                "input_ids": feature["input_ids"] + inputs_remainder,
                "attention_mask": (
                    feature["attention_mask"] + attention_masks_remainder if include_attention_mask else None
                ),
                "labels": (feature["labels"] + labels_remainder),
            }
        elif padding_side == "right":
            feature = {
                "input_ids": np.concatenate([feature["input_ids"], inputs_remainder]).astype(np.int64),
                "attention_mask": (
                    np.concatenate([feature["attention_mask"], attention_masks_remainder]).astype(np.int64)
                    if include_attention_mask
                    else None
                ),
                "labels": (
                    np.concatenate([feature["labels"], labels_remainder]).astype(np.int64) if include_labels else None
                ),
            }
        else:
            feature = {
                "input_ids": np.concatenate([inputs_remainder, feature["input_ids"]]).astype(np.int64),
                "attention_mask": (
                    np.concatenate([attention_masks_remainder, feature["attention_mask"]]).astype(np.int64)
                    if include_attention_mask
                    else None
                ),
                "labels": (
                    np.concatenate([labels_remainder, feature["labels"]]).astype(np.int64) if include_labels else None
                ),
            }

        inputs.append(feature["input_ids"])
        if include_attention_mask:
            attention_masks.append(feature["attention_mask"])

        if include_labels:
            labels.append(feature["labels"])

    output_dict = {"input_ids": torch.from_numpy(np.stack(inputs))}

    if include_attention_mask:
        output_dict["attention_mask"] = torch.from_numpy(np.stack(attention_masks))

    if include_labels:
        output_dict["labels"] = torch.from_numpy(np.stack(labels))

    return output_dict


# function derived from 'transformers' library: https://github.com/huggingface/transformers/blob/main/src/transformers/data/data_collator.py#L52
def pad_without_fast_tokenizer_warning(tokenizer, *pad_args, **pad_kwargs):
    """
    Pads without triggering the warning about how using the pad function is sub-optimal when using a fast tokenizer.
    """

    # To avoid errors when using Feature extractors
    if not hasattr(tokenizer, "deprecation_warnings"):
        return tokenizer.pad(*pad_args, **pad_kwargs)

    # Save the state of the warning, then disable it
    warning_state = tokenizer.deprecation_warnings.get("Asking-to-pad-a-fast-tokenizer", False)
    tokenizer.deprecation_warnings["Asking-to-pad-a-fast-tokenizer"] = True

    try:
        padded = tokenizer.pad(*pad_args, **pad_kwargs)
    finally:
        # Restore the state of the warning.
        tokenizer.deprecation_warnings["Asking-to-pad-a-fast-tokenizer"] = warning_state

    return padded


def stack_tensor_dict(tensor_dicts: list[dict[torch.Tensor]]):
    keys = tensor_dicts[0].keys()
    return {key: torch.stack([d[key] for d in tensor_dicts]) for key in keys}


def stack_iterables(iterables: list[list | tuple]):
    return torch.stack([torch.tensor(iterable) for iterable in iterables])


class DataCollatorForSeq2Seq:
    """
    Automatically adds efficient padding for 'inputs', 'attention_mask' and 'labels'.
    This works for multiple inputs from the dataset logic. If any of the objects does not
    correspond to a dictionary-like structure of a decoded tokenizer's output, it will
    apply the default collate function derived from PyTorch.

    The output of a dictionary-like with key 'input_ids' will have the following keys:
        - `input_ids`
        - `attention_mask` (if found)
        - `labels` (if found)

    This implementation derives from `transformers` library:
    https://github.com/huggingface/transformers/blob/main/src/transformers/data/data_collator.py#L543

    Args:
        tokenizer (`Any`):
            Tokenizer using HuggingFace standard.
        label_pad_token_id (`int`, *optional*, defaults to `-100`):
            Label pad token id. Labels with this value will be ignored in the training process.
    """

    def __init__(self, tokenizer: Any, label_pad_token_id: int = -100):
        self.tokenizer = tokenizer
        self.pad_token_id = self.tokenizer.pad_token_id
        self.label_pad_token_id = label_pad_token_id
        self.padding_side = self.tokenizer.padding_side

    def _collate_nested_dicts(self, data):
        elem = data[0]
        if isinstance(elem, (dict, BatchEncoding)) and "input_ids" in elem:
            return collate_tokenizer_inputs(
                data,
                pad_token_id=self.pad_token_id,
                label_pad_token_id=self.label_pad_token_id,
                padding_side=self.padding_side,
            )
        elif isinstance(elem, (dict, BatchEncoding)) and "input_ids" not in elem:
            return {key: self._collate_nested_dicts([d[key] for d in data]) for key in elem}

        return default_collate(data)

    def _collate_multiple(self, batch: list, length_elems: int) -> Union[tuple, Any]:
        stacked_elems = [[] for _ in range(length_elems)]
        for elem in batch:
            for elem_index in range(length_elems):
                stacked_elems[elem_index].append(elem[elem_index])

        del batch
        for elem_index in range(length_elems):
            elem = stacked_elems[elem_index][0]
            if isinstance(elem, (dict, BatchEncoding)) and "input_ids" in elem:
                stacked_elems[elem_index] = collate_tokenizer_inputs(
                    stacked_elems[elem_index],
                    pad_token_id=self.pad_token_id,
                    label_pad_token_id=self.label_pad_token_id,
                    padding_side=self.padding_side,
                )
            elif isinstance(elem, dict):
                stacked_elems[elem_index] = self._collate_nested_dicts(stacked_elems[elem_index])
            else:
                stacked_elems[elem_index] = default_collate(stacked_elems[elem_index])

        return tuple(stacked_elems) if length_elems > 1 else stacked_elems[0]

    def __call__(self, batch: list) -> Union[tuple, Any]:
        length_elems = len(batch[0]) if isinstance(batch[0], tuple) else 1

        if length_elems == 1:
            if isinstance(batch[0], BatchEncoding) or (isinstance(batch[0], dict) and "input_ids" in batch[0]):
                return collate_tokenizer_inputs(
                    batch,
                    pad_token_id=self.pad_token_id,
                    label_pad_token_id=self.label_pad_token_id,
                    padding_side=self.padding_side,
                )
            else:
                return default_collate(batch)

        return self._collate_multiple(batch, length_elems)


class DataCollatorForLongestSequence:
    """
    Automatically adds efficient padding for inputs, while preserving static labels.

    If output of `__getitem__` Dataset logic looks like:
        `return x, y` (x being a dictionary containing keys `input_ids` and `attention_mask`)
    then the output of the collator function will be `(x, y)`, `x` being the padded inputs with
    the same keys and `y` the stacked labels.

    If output of `__getitem__` Dataset logic looks like:
        `return x` (x being a dictionary containing keys `input_ids` and `attention_mask`)
    then the output of the collator function will be `x`, being the padded inputs with the same keys.

    NOTE: This collator should be used when labels on your dataset logic are not sequences. If that's the case,
    see `DataCollatorForSeq2Seq`.

    Args:
        tokenizer (`Any`):
            Tokenizer using HuggingFace standard.
    """

    def __init__(self, tokenizer: Any, torch_stack: bool = True):
        self.tokenizer = tokenizer
        self.pad_token_id = self.tokenizer.pad_token_id
        self.padding_side = self.tokenizer.padding_side
        self.torch_stack = torch_stack
        self.device = None

    def __call__(self, batch: list):
        inputs = []
        for feature in batch:
            # if feature is a tuple, then it would be of type (inputs, targets)
            if isinstance(feature, tuple):
                feature = feature[0]  # just take first element
            inputs.append(len(feature["input_ids"]))

        max_input_length = max(inputs)

        inputs = []
        attention_masks = []
        labels = []
        for feature in batch:
            if isinstance(feature, tuple):
                labels.append(feature[1])
                feature = feature[0]
            inputs_remainder = [self.pad_token_id] * (max_input_length - len(feature["input_ids"]))
            attention_masks_remainder = [0] * (max_input_length - len(feature["input_ids"]))

            if self.padding_side == "right":
                feature = {
                    "input_ids": np.concatenate([feature["input_ids"], inputs_remainder]).astype(np.int64),
                    "attention_mask": np.concatenate([feature["attention_mask"], attention_masks_remainder]).astype(
                        np.int64
                    ),
                }
            else:
                feature = {
                    "input_ids": np.concatenate([inputs_remainder, feature["input_ids"]]).astype(np.int64),
                    "attention_mask": np.concatenate([attention_masks_remainder, feature["attention_mask"]]).astype(
                        np.int64
                    ),
                }

            inputs.append(feature["input_ids"])
            attention_masks.append(feature["attention_mask"])

        output = {
            "input_ids": torch.from_numpy(np.stack(inputs)),
            "attention_mask": torch.from_numpy(np.stack(attention_masks)),
        }

        if len(labels) > 0:
            if isinstance(labels[0], dict):
                keys = labels[0].keys()
                if self.torch_stack:
                    out_labels = {k: torch.stack([label[k] for label in labels]) for k in keys}
                else:
                    out_labels = {k: [label[k] for label in labels] for k in keys}
            elif isinstance(labels[0], torch.Tensor):
                out_labels = labels
                if self.torch_stack:
                    out_labels = torch.stack(out_labels)
            elif isinstance(labels[0], (list, tuple, np.ndarray)):
                out_labels = [torch.tensor(label, device=self.device) for label in labels]
                if self.torch_stack:
                    out_labels = torch.stack(out_labels)
            else:
                out_labels = None

            return output, out_labels

        return output


class DataCollatorForLanguageModeling:
    """
    Collator function to implement automatic language modeling, such as
    Masked Language Modeling.

    Args:
        tokenizer (`Any`):
            Tokenizer using HuggingFace standard.
        mlm (`bool`, *optional*, defaults to `True`):
            Implements Masked Language Modeling.
        mlm_probability (`float`, *optional*, defaults to `0.15`):
            How much masking is implemented in Masked Language Modeling.
        ignore_index (`int`, *optional*, defaults to `-100`):
            Label pad token id. Labels with this value will be ignored in the training process.
        masked_to_mask (`float`, *optional*, defaults to `0.8`):
            Probability to replace masked input tokens with mask token. The half remaining percent will
            replace masked input tokens with random word, and the other half will keep the masked input tokens
            unchanged. If `apply_random_words` is set to `False`, then the entire remaining percent will be unchanged.
        apply_random_words (`bool`, *optional*, defaults to `True`):
            Whether to apply random words during Masked Language Modeling.
        force_one_output (`bool`, *optional*, defaults to `False`):
            Whether to force output one output. If Dataset object `__getitem__` function returns a tuple, only the first
            element will be considered and extra targets will be dropped.
    """

    def __init__(
        self,
        tokenizer: Any,
        mlm: bool = True,
        mlm_probability: float = 0.15,
        ignore_index: int = -100,
        masked_to_mask: float = 0.8,
        apply_random_words: bool = True,
        force_one_output: bool = False,
    ) -> Union[dict, tuple[dict, torch.Tensor]]:
        self.tokenizer = tokenizer
        self.mlm = mlm
        self.mlm_probability = mlm_probability
        self.ignore_index = ignore_index
        self.masked_to_mask = masked_to_mask
        self.apply_random_words = apply_random_words
        self.force_one_output = force_one_output

    def __call__(self, batch: list) -> dict:
        has_extra_targets = isinstance(batch[0], (tuple, list))
        if not has_extra_targets:
            tokenizer_dict = pad_without_fast_tokenizer_warning(self.tokenizer, batch, return_tensors="pt")
        else:
            tokenizer_dict_batch, extra_targets = [], []
            for elems in batch:
                tokenizer_dict_batch.append(elems[0])
                if not self.force_one_output:
                    extra_targets.append(elems[1:])

            tokenizer_dict = pad_without_fast_tokenizer_warning(
                self.tokenizer, tokenizer_dict_batch, return_tensors="pt"
            )

        special_tokens_mask = tokenizer_dict.pop("special_tokens_mask", None)
        if self.mlm:
            tokenizer_dict["input_ids"], tokenizer_dict["labels"] = self.torch_mask_tokens(
                tokenizer_dict["input_ids"], special_tokens_mask=special_tokens_mask
            )
        else:
            labels = tokenizer_dict["input_ids"].clone()
            if self.tokenizer.pad_token_id is not None:
                labels[labels == self.tokenizer.pad_token_id] = self.ignore_index
            tokenizer_dict["labels"] = labels

        if has_extra_targets and not self.force_one_output:
            num_elems = len(extra_targets[0])
            extra_targets_return = [[] for _ in range(num_elems)]

            for target in extra_targets:
                for idx in range(num_elems):
                    tgt = target[idx]
                    extra_targets_return[idx].append(tgt)

            stack_funcs = []
            for idx, extra_target_return in enumerate(extra_targets_return):
                first_elem = extra_target_return[0]
                if isinstance(first_elem, torch.Tensor):
                    stack_funcs.append(torch.stack)
                elif isinstance(first_elem, dict):
                    stack_funcs.append(stack_tensor_dict)
                elif isinstance(first_elem, (tuple, list)):
                    stack_funcs.append(stack_iterables)
                else:
                    stack_funcs.append(torch.tensor)

            extra_targets_return = [
                stack_funcs[idx](extra_target_return) for idx, extra_target_return in enumerate(extra_targets_return)
            ]

        if has_extra_targets and not self.force_one_output:
            return tokenizer_dict, *extra_targets_return

        return tokenizer_dict

    def torch_mask_tokens(self, inputs: torch.Tensor, special_tokens_mask):
        labels = inputs.clone()

        probability_matrix = torch.full(labels.shape, self.mlm_probability)
        if special_tokens_mask is None:
            special_tokens_mask = [
                self.tokenizer.get_special_tokens_mask(val, already_has_special_tokens=True) for val in labels.tolist()
            ]
            special_tokens_mask = torch.tensor(special_tokens_mask, dtype=torch.bool)
        else:
            special_tokens_mask = special_tokens_mask.bool()

        probability_matrix.masked_fill_(special_tokens_mask, value=0.0)
        masked_indices = torch.bernoulli(probability_matrix).bool()
        labels[~masked_indices] = self.ignore_index

        indices_replaced = torch.bernoulli(torch.full(labels.shape, self.masked_to_mask)).bool() & masked_indices
        inputs[indices_replaced] = self.tokenizer.convert_tokens_to_ids(self.tokenizer.mask_token)

        if self.apply_random_words:
            indices_random = torch.bernoulli(torch.full(labels.shape, 0.5)).bool() & masked_indices & ~indices_replaced
            random_words = torch.randint(len(self.tokenizer), labels.shape, dtype=torch.long)
            inputs[indices_random] = random_words[indices_random]

        return inputs, labels



# Copyright 2025 ghanvert. All rights reserved.
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#     http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.

from typing import Any, Optional, Union

import torch
import torch.distributed as dist
import torch.nn.functional as F

from .utility import WORLD_SIZE


def pad_to(tensor: torch.Tensor, maximum: int) -> tuple[torch.Tensor, torch.Tensor]:
    """Pad on first dimension."""
    target_rows = maximum

    current_rows, _ = tensor.size()
    rows_to_add = target_rows - current_rows

    if rows_to_add > 0:
        padded_tensor = F.pad(tensor, (0, 0, 0, rows_to_add), mode="constant", value=0)
        padded_tensor[-rows_to_add:] = tensor[-1]
    else:
        padded_tensor = tensor[:target_rows]

    padded_mask = torch.tensor(
        [i < len(tensor) for i in range(len(padded_tensor))], dtype=torch.long, device=padded_tensor.device
    )
    return padded_tensor, padded_mask


def drop_duplicates(tensor: torch.Tensor, padded_mask: torch.Tensor = None) -> torch.Tensor:
    return tensor[padded_mask.bool()]


def gather(tensor: torch.Tensor, num_processes: int = None) -> torch.Tensor:
    if WORLD_SIZE == 1:
        return tensor

    if num_processes is None:
        num_processes = WORLD_SIZE

    tensor_ = tensor.clone()  # TODO does this still apply??
    collected = [torch.empty(tensor.shape, dtype=tensor.dtype, device=tensor.device) for _ in range(num_processes)]
    dist.all_gather(collected, tensor_)
    collected = torch.cat(collected)

    return collected


def gather_object(obj: Any, num_processes: int = None) -> list[Any]:
    if WORLD_SIZE == 1:
        return obj

    if num_processes is None:
        num_processes = WORLD_SIZE

    collected = [None] * num_processes
    dist.all_gather_object(collected, obj)

    return collected


def gather_into_single_process(tensor: Optional[torch.Tensor], dst: int = 0, remainder: int = 0) -> torch.Tensor:
    if WORLD_SIZE == 1:
        return tensor

    if tensor is not None and tensor.ndimension() == 0:
        tensor = tensor.unsqueeze(0)

    world_size = dist.get_world_size()
    rank = dist.get_rank()
    collected = [None for _ in range(world_size)]
    dist.gather_object(tensor, collected if rank == dst else None, dst=dst)
    output = None
    if rank == dst:
        rank_device = f"cuda:{dst}"
        if remainder > 0:
            collected = collected[:remainder]
        tensors = [tensor.to(rank_device) for tensor in collected if tensor is not None]
        if len(tensors) > 0:
            output = torch.cat(tensors)

    return output


def gather_and_drop_duplicates(tensor: torch.Tensor, maximum: int) -> torch.Tensor:
    tensor, padding_mask = pad_to(tensor, maximum)

    # gather operations
    tensor = gather(tensor)
    padding_mask = gather(padding_mask)

    tensor = drop_duplicates(tensor, padding_mask)

    return tensor


def unique_gather(obj: Union[torch.Tensor, Any], remainder: int = 0) -> Union[torch.Tensor, Any]:
    obj = gather(obj) if isinstance(obj, torch.Tensor) else gather_object(obj)

    if remainder == 0:
        return obj

    true_size = len(obj) - remainder
    return obj[:true_size]



# Copyright 2025 ghanvert. All rights reserved.
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#     http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.

import inspect
import math
import os
import traceback
from typing import Optional, Union

import torch
import torch.nn as nn
from accelerate import DistributedType
from accelerate.utils import LoggerType, ProjectConfiguration, set_seed, tqdm
from torch.utils.data import DataLoader, Dataset
from typing_extensions import Any

from .callbacks import Callback
from .dataloader_samplers import BaseSampler
from .dist_utils import gather_into_single_process
from .handlers import Handler
from .hyperparameters import HyperParameters
from .metrics import Metric
from .modules import AcceleratorModule
from .monitor import Monitor
from .states import TrainingState
from .utils import (
    cleanup,
    combine_dicts,
    get_number_and_unit,
    is_url,
    operator_map,
    read_status,
    save_status,
    time_prefix,
)


CHECKPOINT_PATH = "checkpoint"
STATUS_PATH = "status.json"
DEBUG_MODE: int = int(os.environ.get("ACCMT_DEBUG_MODE", 0))


class Trainer:
    """
    Class to implement the training configuration.
    """

    @classmethod
    def from_config(
        cls,
        config: Union[str, dict],
        log_with: Optional[Union[Any, list]] = None,
        sampler: Optional[Union[Any, list]] = None,
        collate_fn_train: Optional[Any] = None,
        collate_fn_val: Optional[Any] = None,
    ):
        """
        Load a configuration from a file or a dictionary.

        Args:
            config (`str` or `dict`):
                Path to a file or dictionary containing kwargs for Trainer constructor. The file can
                be YAML or JSON.
            log_with (`accmt.tracker` or `list`, *optional*, defaults to `None`):
                Logger to log metrics.
            sampler (list or `Any`, *optional*, defaults to `None`):
                Sampler (or list of samplers) for train DataLoader.
            collate_fn_train (`function` or `list`, *optional*, defaults to `None`):
                Collate function to be implemented in train dataloader.
            collate_fn_val (`function` or `list`, *optional*, defaults to `None`):
                Collate function to be implemented in validation dataloader.
        """
        assert isinstance(config, (str, dict)), "'config' needs to be either a path to a file, or a dictionary."
        if isinstance(config, str):
            import yaml

            config = yaml.safe_load(open(config))

        return Trainer(
            **config,
            log_with=log_with,
            sampler=sampler,
            collate_fn_train=collate_fn_train,
            collate_fn_val=collate_fn_val,
        )

    def __init__(
        self,
        hps_config: Union[str, dict, HyperParameters],
        model_path: str,
        track_name: Optional[str] = None,
        checkpoint: Optional[str] = None,
        resume: Optional[bool] = None,
        model_saving: Optional[Union[str, list[str]]] = "best_valid_loss",
        evaluate_every_n_steps: Optional[int] = None,
        checkpoint_every: Optional[str] = "epoch",
        logging_dir: str = "logs",
        log_with: Optional[Union[Any, list]] = None,
        log_every: Optional[int] = 1,
        grad_accumulation_steps: Optional[int] = None,
        clip_grad: Optional[float] = None,
        set_to_none: bool = True,
        shuffle_train: bool = True,
        shuffle_validation: bool = False,
        sampler: Optional[Union[Any, list]] = None,
        model_saving_below: Optional[float] = None,
        model_saving_above: Optional[float] = None,
        collate_fn_train: Optional[Any] = None,
        collate_fn_val: Optional[Any] = None,
        disable_collate_fn_on_evaluation: bool = False,
        max_shard_size: str = "10GB",
        safe_serialization: bool = False,
        compile: bool = False,
        train_loss_metric_name: str = "train_loss",
        val_loss_metric_name: str = "val_loss",
        dataloader_pin_memory: bool = True,
        dataloader_num_workers: Optional[int] = None,
        dataloader_drop_last: bool = False,
        handlers: Optional[Union[list, Any]] = None,
        eval_when_finish: bool = True,
        eval_when_start: bool = False,
        monitor: Optional[Monitor] = None,
        metrics: Optional[Union[Metric, list[Metric]]] = None,
        cleanup_cache_every_n_steps: Optional[int] = None,
        callback: Optional[Callback] = None,
        gather_none: bool = False,
        additional_tracker_config: Optional[dict[str, Any]] = None,
        report_train_loss_per_epoch: bool = False,
        **kwargs: Optional[Any],
    ):
        """
        Trainer constructor to set configuration.

        Args:
            hps_config (`str`, `dict`, or `HyperParameters`):
                YAML hyperparameters file path, dictionary or `HyperParameters`.
            model_path (`str`):
                Path to save model.
            track_name (`str`, *optional*, defaults to `None`):
                Track name for trackers. If set to `None` (default), the track name will be
                the model's folder name.
            checkpoint (`str`, *optional*, defaults to `None`):
                Path where to save the checkpoint. Path by default is going to be of type:
                'checkpoint-MODEL_PATH_NAME'.
            resume (`bool`, *optional*, defaults to `None`):
                Whether to resume from checkpoint. Default option is `None`, which means resuming from checkpoint
                will be handled automatically, whether the checkpoint directory exists or not.
            model_saving (`str` or `list`, *optional*, defaults to `best_valid_loss`):
                Type of model saving. It can be one of the following values:

                - `"best_valid_loss"`: Saves the model whenever the validation loss is the best recorded.
                - `"best_train_loss"`: Saves the model whenever the training loss is the best recorded.
                - `"always"`: Saves the model always at the end of every evaluation.
                - in format of `"best_{METRIC}"`, where METRIC corresponds to the additional metric in `additional_metrics`.

                If not specified (`None`), model saving will be disabled. This can also be a list of model savings methods to save
                more models on different metrics. When implementing multiple model savings, the resulting model path will be of the form
                "{model_path}_best_{metric}", like the following examples:
                    - MODEL_best_accuracy
                    - MODEL_best_train_loss
            evaluate_every_n_steps (`int`, *optional*, defaults to `None`):
                Evaluate model in validation dataset (if implemented) every N steps. If this is set
                to `None` (default option), evaluation will happen at the end of every epoch.
            checkpoint_every (`str`, *optional*, defaults to `epoch`):
                Checkpoint every N epochs, steps or evaluations. Requires a number and a unit in a string.
                The following examples are valid:

                - `"epoch"`, `"ep"`, `"1epoch"`, `"1ep"`, `"1 epoch"`, `"1 ep"`: 1 Epoch
                - `"step"`, `"st"`, `"1step"`, `"1st"`, `"1 step"`, `"1 st"`: 1 Step
                - `"evaluation"`, `"eval"`, `"1evaluation"`, `"1eval"`, `"1 evaluation"`, `"1 eval"`: 1 Evaluation

                (a character `s` at the end of the string is also valid)

                If set to `None`, checkpointing will be disabled.
            logging_dir (`str`, *optional*, defaults to `logs`):
                Path where to save logs to show progress. It can be an IP address (local or remote), HTTP or HTTPS link,
                or simply a directory.
            log_with (`accmt.tracker` or `list`, *optional*, defaults to `None`):
                Logger to log metrics. It can be one of the following imports from accmt:

                    - `TensorBoard`
                    - `WandB`
                    - `CometML`
                    - `Aim`
                    - `MLFlow`
                    - `ClearML`
                    - `DVCLive`
            log_every (`int`, *optional*, defaults to `1`):
                Log every N steps. If `grad_accumulation_steps` is set to a higher value than `1`, then this parameter will be
                modified to be `log_every` * `grad_accumulation_steps` ONLY when logging during training.
            grad_accumulation_steps (`int`, *optional*, defaults to `None`):
                Accumulate gradients for N steps. Useful for training large models and simulate
                large batches when memory is not enough. If set to `None` or `1`, no accumulation will be perfomed.
            clip_grad (`float`, *optional*, defaults to `None`):
                Performs gradient clipping in between backpropagation and optimizer's step function. This feature is disabled when
                using DeepSpeed, because it handles gradient clipping in the configuration file. If you wan't to configure gradient
                clipping, you might want to use Accelerate's CLI to create a new config file.
            set_to_none (`bool`, *optional*, defaults to `True`):
                From PyTorch documentation: "instead of setting to zero, set the grads to None. This will
                in general have lower memory footprint, and can modestly improve performance." Some
                optimizers have a different behaviour if the gradient is 0 or None. See PyTorch docs
                for more information: https://pytorch.org/docs/stable/generated/torch.optim.Optimizer.zero_grad.html
            shuffle_train (`bool`, *optional*, defaults to `True`):
                Whether to shuffle train DataLoader.
            shuffle_validation (`bool`, *optional*, defaults to `False`):
                Whether to shuffle validation DataLoader.
            sampler (`list` or `Any`, *optional*, defaults to `None`):
                Sampler (or list of samplers) for train DataLoader.
            model_saving_below (`float`, *optional*, defaults to `None`):
                Start saving model below this metric (based on `model_saving`).
            model_saving_above (`float`, *optional*, defaults to `None`):
                Start saving model above this metric (based on `model_saving`).
            collate_fn_train (`function` or `list`, *optional*, defaults to `None`):
                Collate function to be implemented in train dataloader. If `module` overrides `collate_fn` from
                `AcceleratorModule` class, then that function will be used instead of the one specified on
                this constructor. If a list of collate functions is given, then the every collate function will affect
                the batch in the given order.
            collate_fn_val (`function` or `list`, *optional*, defaults to `None`):
                Collate function to be implemented in validation dataloader.
            disable_collate_fn_on_evaluation (`bool`, *optional*, defaults to `False`):
                Disable 'collate_fn' on validation DataLoader.
            max_shard_size (`str`, *optional*, defaults to `10GB`):
                Max model shard size to be used.
            safe_serialization (`bool`, *optional*, defaults to `False`):
                Whether to save model using safe tensors or the traditional PyTorch way. If `True`, some tensors
                will be lost.
            compile (`bool`, *optional*, defaults to `False`):
                Whether to call `torch.compile` on model (and teacher, if implemented).
            train_loss_metric_name (`str`, *optional*, defaults to `train_loss`):
                Metric name for train loss in logs.
            val_loss_metric_name (`str`, *optional*, defaults to `val_loss`):
                Metric name for validation loss in logs.
            dataloader_pin_memory (`bool`, *optional*, defaults to `True`):
                Enables pin memory option in DataLoader.
            dataloader_num_workers (`int`, *optional*, defaults to `None`):
                Number of processes for DataLoader. This defaults to `None`, meaning the number of workers will be equal to the
                number of processes set for training.
            dataloader_drop_last (`bool`, *optional*, defaults to `False`):
                Whether to drop last batch on DataLoader or not.
            handlers (`Any` or `list`, *optional*, defaults to `None`):
                Handler or List of handlers to catch errors and make a safe checkpoint.
            eval_when_finish (`bool`, *optional*, defaults to `True`):
                At the end of training, evaluate model on validation dataset (if available). This option is only valid when
                `evaluate_every_n_steps` is not `None`.
            eval_when_start (`bool`, *optional*, defaults to `False`):
                Start training with evaluation (if available).
            monitor (`Monitor` or `dict`, *optional*, defaults to `None`):
                Monitor arguments to keep track of variables during training. If not specified, 'train_loss' and 'validation_loss' will
                be set to `True` by default.

                NOTE: Learning rate, GPU and CPU monitoring will only be reported during training, not evaluation. Also, GPU and CPU
                monitoring will only be reported on main process (index 0).
            metrics (`list`, *optional*, defaults to `None`):
                List of additional metrics of type 'Metric' to track.
            cleanup_cache_every_n_steps (`int`, *optional*, defaults to `None`):
                Cleanup CPU and CUDA caches every N steps. Default is no cleanup.

                NOTE: Every epoch and every evaluation call we cleanup cache.
            callback (`Callback`, *optional*, defaults to `None`):
                `Callback` to implement. This module will have the logic for every existing callback function.
            gather_none (`bool`, *optional*, defaults to `False`):
                In evaluation, allow `None` objects across different processes to be collected. This incurs in
                GPU -> CPU transfer. By default (`False`), tensors are gathered considering that there are no `None` or any
                object types, only `Tensor` objects.
            additional_tracker_config (`dict`, *optional*, defaults to `None`):
                Additional configuration specification for tracker (e.g. hyper-parameters).
            report_train_loss_per_epoch (`bool`, *optional*, defaults to `False`):
                Report train loss at the end of every epoch (instead of `log_every`).
            kwargs (`Any`, *optional*):
                Extra arguments for specific `init` function in Tracker, e.g. `run_name`, `tags`, etc.
        """
        assert isinstance(hps_config, (str, dict, HyperParameters)), (
            "'hps_config' needs to be either a string, dictionary or HyperParameters class."
        )
        from . import accelerator

        self.accelerator = accelerator
        self.hps = HyperParameters.from_config(hps_config) if isinstance(hps_config, (str, dict)) else hps_config
        self.track_name = track_name
        self.checkpoint = checkpoint if checkpoint is not None else "checkpoint"
        self.checkpoint = f"{model_path}/{self.checkpoint}"
        self.resume = (
            resume if resume is not None else os.path.exists(self.checkpoint) and len(os.listdir(self.checkpoint)) > 0
        )
        if DEBUG_MODE >= 3:
            self.resume = False
        self.model_path = model_path
        self.metrics = metrics if isinstance(metrics, list) else [metrics]
        self.metrics = [metric for metric in self.metrics if metric is not None]

        _default_model_savings = set({"best_valid_loss", "best_train_loss", "always"})
        _implemented_metrics = set(
            f"best_{metric.main_metric}" if not metric.main_metric.startswith("best_") else metric
            for metric in self.metrics
        )
        _implemented_metrics.update(_default_model_savings)
        self.model_saving = model_saving if isinstance(model_saving, list) else [model_saving]
        self.model_saving = [ms.lower() for ms in self.model_saving]
        self.model_saving = [f"best_{ms}" if not ms.startswith("best_") else ms for ms in self.model_saving]
        assert all(ms in _implemented_metrics for ms in self.model_saving), (
            f"All 'model_saving' methods should be declared in 'metrics' or be one of {_default_model_savings}."
        )

        self.evaluate_every_n_steps = evaluate_every_n_steps
        self.checkpoint_every = checkpoint_every
        if self.checkpoint_every is not None:
            self.checkpoint_every, self.checkpoint_strat = get_number_and_unit(self.checkpoint_every)
            self.enable_checkpointing = True
        else:
            # fix invalid arguments
            self.checkpoint_every = 1
            self.checkpoint_strat = "epoch"
            self.enable_checkpointing = False

        self.logging_dir = logging_dir
        self.log_with = None
        self.log_every = log_every
        self.grad_accumulation_steps = grad_accumulation_steps if grad_accumulation_steps is not None else 1
        assert clip_grad is None or isinstance(clip_grad, float), "'clip_grad' argument needs to be a float."
        if clip_grad is not None and self.accelerator.distributed_type == DistributedType.DEEPSPEED:
            self.accelerator.print(
                time_prefix(),
                "[WARNING] Clipping gradient using Trainer is not supported when running with DeepSpeed. Setting it to None.",
            )
            clip_grad = None
        self.clip_grad = clip_grad
        self.set_to_none = set_to_none
        self.shuffle_train = shuffle_train
        self.shuffle_validation = shuffle_validation
        self.sampler = sampler
        self.model_saving_below = model_saving_below if model_saving_below is not None else float("inf")
        self.model_saving_above = model_saving_above if model_saving_above is not None else float("-inf")
        self.collate_fn_train = (
            self._get_collate_fn_pipeline(collate_fn_train) if isinstance(collate_fn_train, list) else collate_fn_train
        )
        self.collate_fn_val = (
            self._get_collate_fn_pipeline(collate_fn_val) if isinstance(collate_fn_val, list) else collate_fn_val
        )
        self.disable_collate_fn_on_evaluation = disable_collate_fn_on_evaluation
        self.max_shard_size = max_shard_size
        self.safe_serialization = safe_serialization
        self.compile = compile
        self.train_loss_metric_name = train_loss_metric_name
        self.val_loss_metric_name = val_loss_metric_name
        self.dataloader_pin_memory = dataloader_pin_memory
        self.dataloader_num_workers = (
            dataloader_num_workers if dataloader_num_workers is not None else self.accelerator.num_processes
        )
        self.dataloader_drop_last = dataloader_drop_last
        self.handlers = handlers if isinstance(handlers, list) else [handlers]
        if self.handlers[0] is not None:
            raise NotImplementedError("'handlers' argument is not yet fully implemented.")
        self.eval_when_finish = eval_when_finish
        self.eval_when_start = eval_when_start if DEBUG_MODE < 4 else False
        self.monitor = monitor if isinstance(monitor, Monitor) else Monitor.from_config(monitor)
        self.monitor._debug_mode = DEBUG_MODE
        self.monitor.grad_norm = (
            self.monitor.grad_norm if self.accelerator.distributed_type == DistributedType.DEEPSPEED else False
        )
        if self.monitor.grad_norm and self.accelerator.distributed_type == DistributedType.DEEPSPEED:
            self.accelerator.print(
                time_prefix(),
                "[WARNING] Gradient norm monitoring is not yet supported when running with DeepSpeed. Setting it to False.",
            )
            self.monitor.grad_norm = False
        self.cleanup_cache_every_n_steps = cleanup_cache_every_n_steps
        self.callback = callback if callback is not None else Callback()
        self.gather_none = gather_none
        self.additional_tracker_config = additional_tracker_config if additional_tracker_config is not None else {}
        self.init_kwargs = kwargs

        self.accelerator.project_configuration = ProjectConfiguration(
            project_dir=".", logging_dir=logging_dir, total_limit=1
        )

        if log_with is not None:
            if not isinstance(log_with, list):
                log_with = [log_with]
            self.log_with = [tracker for tracker in log_with]
            if DEBUG_MODE < 1:
                self.accelerator.log_with = [tracker.tracker for tracker in log_with]

        self.report_train_loss_per_epoch = report_train_loss_per_epoch

        # we need to calculate mean for these tensors.
        self.train_total_loss: torch.Tensor = None  # loss tensor for evaluation
        self.train_track_loss: torch.Tensor = None  # train loss tensor to be reported
        self.val_total_loss: torch.Tensor = None  # val loss tensor for evaluation
        self.val_track_loss: torch.Tensor = None  # val loss tensor to be reported

        self.train_dataloader: DataLoader = None
        self.val_dataloader: DataLoader = None

        self.module = None

    def fit(
        self,
        module: Union[AcceleratorModule, str, Union[tuple[str, str], tuple[str, Any]]],
        train_dataset: Optional[Dataset] = None,
        val_dataset: Optional[Dataset] = None,
        **kwargs: Any,
    ):
        """
        Function to train a given `AcceleratorModule`.

        Args:
            module (`AcceleratorModule`, `str` or `tuple`):
                `AcceleratorModule` class containig the training logic. This can also be a string specifying a
                HuggingFace model, or a tuple of type (model, type), where 'model' is a string for the HuggingFace model,
                and 'type' is a string or class (from transformers library) for the model type.
            train_dataset (`torch.utils.data.Dataset`, *optional*, defaults to `None`):
                `Dataset` class from PyTorch containing the train dataset logic. If not provided, then
                `get_train_dataloader` from `module` will be used to get the train DataLoader.
            val_dataset (`torch.utils.data.Dataset`, *optional*, defaults to `None`):
                `Dataset` class from PyTorch containing the validation dataset logic. If this
                dataset is not specified, then the validation logic of `AcceleratorModule`
                (if specified) will be skipped. If `model_saving` parameter in the constructor is set
                to `best_valid_loss`, this will be converted to `best_train_loss` in the background.
                If not provided, it will use `get_validation_dataloader` to get the validation DataLoader
                (if implemented).
            kwargs (`Any`):
                Keyword arguments for `from_pretrained` function for model initialization.
        """
        self.module = module
        self.monitor.additional_metrics = self.metrics is not None and len(self.metrics) > 0
        self.monitor._do_tracking = self.log_with is not None
        self.train_total_loss = torch.tensor(0.0, device=self.accelerator.device, pin_memory=True)
        self.train_track_loss = torch.tensor(0.0, device=self.accelerator.device, pin_memory=True)
        self.val_total_loss = torch.tensor(0.0, device=self.accelerator.device, pin_memory=True)
        self.val_track_loss = torch.tensor(0.0, device=self.accelerator.device, pin_memory=True)

        if isinstance(module, str):
            module = AcceleratorModule.from_hf(module, **kwargs)
        elif isinstance(module, tuple):
            module = AcceleratorModule.from_hf(*module, **kwargs)

        model = getattr(module, "model", None)
        if model is None:
            raise AttributeError("'self.model' needs to be declared in the AcceleratorModule class.")
        elif model is not None and not isinstance(model, nn.Module):
            raise ValueError("'self.model' needs to be an instance of 'nn.Module'.")

        teacher = getattr(module, "teacher", None)
        if teacher is not None and not isinstance(teacher, nn.Module):
            raise ValueError("'self.teacher' needs to be an instance of 'nn.Module'.")

        module._log_every = self.log_every
        if torch.cuda.is_available():
            model.to(self.accelerator.device)  # for optimizer to apply fused when available
            if teacher is not None:
                teacher.eval()
                teacher.to(self.accelerator.device)
        if self.compile and DEBUG_MODE < 2:
            model = torch.compile(model)
            if teacher is not None:
                teacher = torch.compile(teacher)

        if self.accelerator.distributed_type == DistributedType.FSDP:
            # preparing model before dataloaders is only supported by FSDP apparently, and this is the
            # recommended setting to prepare training.
            model = self.accelerator.prepare(model)

        if self.accelerator.is_main_process and DEBUG_MODE < 3:
            os.makedirs(self.model_path, exist_ok=True)

        if self.resume:
            status_dict = read_status(f"{self.checkpoint}/{STATUS_PATH}")
            if "evaluations_done" not in status_dict:
                # in case that ACCMT was updated from < 1.1.0 or 1.2.3 version to a higher one,
                # this fixes it.
                status_dict["evaluations_done"] = 0
                status_dict["additional_metrics"] = {}
        else:
            status_dict = {
                "best_train_loss": float("inf"),
                "best_valid_loss": float("inf"),
                "epoch": 0,
                "epoch_step": 0,
                "global_step": 0,
                "eval_global_step": 0,
                "evaluations_done": 0,
                "train_track_loss": 0,
                "additional_metrics": {},
            }
        module.status_dict = status_dict
        self.monitor._set_extra(self.accelerator, status_dict, self.train_loss_metric_name, self.val_loss_metric_name)

        if module._implemented_collate_fn_train:
            self.collate_fn_train = module.collate_fn_train

        if module._implemented_collate_fn_val:
            self.collate_fn_val = module.collate_fn_val

        is_tuple = hasattr(self.hps.batch_size, "__len__")
        module.batch_size = self.hps.batch_size
        train_batch_size = self.hps.batch_size[0] if is_tuple else self.hps.batch_size
        val_batch_size = self.hps.batch_size[1] if is_tuple and len(self.hps.batch_size) > 1 else self.hps.batch_size
        dl_args = {
            "pin_memory": self.dataloader_pin_memory,
            "num_workers": self.dataloader_num_workers,
            "drop_last": self.dataloader_drop_last,
        }

        train_dataloader = module.get_train_dataloader()
        if train_dataset is not None and train_dataloader is None:
            shuffle_train = self.shuffle_train if self.sampler is None else None
            samplers = None
            if isinstance(self.sampler, list):
                samplers = []
                for sampler in self.sampler:
                    if issubclass(sampler.__class__, BaseSampler):
                        samplers.append(sampler(self.accelerator))
                    else:
                        samplers.append(sampler)
            else:
                samplers = (
                    self.sampler(self.accelerator) if issubclass(self.sampler.__class__, BaseSampler) else self.sampler
                )
            train_dataloader = DataLoader(
                train_dataset,
                shuffle=shuffle_train,
                sampler=samplers,
                batch_size=train_batch_size,
                collate_fn=self.collate_fn_train,
                **dl_args,
            )

            collate_fn_val = self.collate_fn_val if not self.disable_collate_fn_on_evaluation else None

            val_dataloader = module.get_validation_dataloader()
            if val_dataset is not None and val_dataloader is None:
                val_dataloader = DataLoader(
                    val_dataset,
                    shuffle=self.shuffle_validation,
                    batch_size=val_batch_size,
                    collate_fn=collate_fn_val,
                    **dl_args,
                )

            # conditionals
            _EVALUATION_EVERY_N_STEPS = (
                all([val_dataloader is not None, hasattr(module, "validation_step")])
                and self.evaluate_every_n_steps is not None
            )
            _CHECKPOINT_EVERY_N_STEPS = self.enable_checkpointing and self.checkpoint_strat == "step"
            _CHECKPOINT_AFTER_EVALUATION = self.enable_checkpointing and self.checkpoint_strat == "eval"
            _CHECKPOINT_WHEN_EPOCH_ENDS = self.enable_checkpointing and self.checkpoint_strat in {"epoch", "eval"}

            if val_dataloader is None and self.model_saving == "best_valid_loss":
                self.model_saving = "best_train_loss"

            optimizer = module.get_optimizer()
            if optimizer is None:
                optimizer = self._get_optimizer(model)

            scheduler = module.get_scheduler(
                optimizer, round(len(train_dataloader) / self.accelerator.num_processes), self.hps.epochs
            )
            if self.hps.scheduler is not None and scheduler is None:
                scheduler = self._get_scheduler(
                    optimizer, -1, round(len(train_dataloader) / self.accelerator.num_processes), self.hps.epochs
                )
                # -1 for last_epoch since Accelerate will take care of recovering the progress

            if self.log_with is not None and DEBUG_MODE < 1:
                self._initialize_trackers()

            unwrapped_model = model
            unwrapped_teacher = teacher
            unwrapped_optimizer = optimizer
            unwrapped_train_dataloader = train_dataloader
            unwrapped_val_dataloader = val_dataloader
            unwrapped_scheduler = scheduler

            if self.accelerator.distributed_type == DistributedType.FSDP:
                train_dataloader, val_dataloader, optimizer, scheduler = self.accelerator.prepare(
                    train_dataloader, val_dataloader, optimizer, scheduler
                )
            else:
                model, train_dataloader, val_dataloader, optimizer, scheduler = self.accelerator.prepare(
                    model, train_dataloader, val_dataloader, optimizer, scheduler
                )

            if self.accelerator.distributed_type != DistributedType.DEEPSPEED and teacher is not None:
                teacher = self.accelerator.prepare_model(teacher)

            if self.accelerator.distributed_type == DistributedType.FSDP:
                module.model = model  # force module.model to be wrapped to not have problems with dimmensions

            # NOTE: we aren't forcing module.model to be wrapped for other distributed types since we haven't seen any
            # issues with training, and it's actually a little bit faster doing inference with the model directly on the Module class.

            self.model = model

            if scheduler is not None:
                self.accelerator.register_for_checkpointing(scheduler)

            if self.log_with is not None:
                track_name = self.model_path.split("/")[-1] if self.track_name is None else self.track_name
                init_kwargs = combine_dicts(*[tracker.init(**self.init_kwargs) for tracker in self.log_with])

                config = self.hps.get_config()
                effective_num = self.grad_accumulation_steps * self.accelerator.num_processes
                config["effective_batch_size"] = (
                    tuple(batch_size * effective_num for batch_size in self.hps.batch_size)
                    if isinstance(self.hps.batch_size, (tuple, list))
                    else self.hps.batch_size * effective_num
                )
                config["grad_accumulation_steps"] = self.grad_accumulation_steps
                config["num_processes"] = self.accelerator.num_processes

                if DEBUG_MODE < 1:
                    tracker_config = config | self.additional_tracker_config
                    self.accelerator.init_trackers(track_name, config=tracker_config, init_kwargs=init_kwargs)

            if self.resume:
                self.callback.on_resume()
                if os.path.exists(self.checkpoint):
                    self.accelerator.load_state(f"{self.checkpoint}/{CHECKPOINT_PATH}")
                else:
                    raise FileNotFoundError(f"{self.checkpoint} was not found.")

        self.train_dataloader = train_dataloader
        self.val_dataloader = val_dataloader
        self.module.state = TrainingState(
            model=unwrapped_model,
            wrapped_model=model,
            teacher=unwrapped_teacher,
            wrapped_teacher=teacher,
            optimizer=unwrapped_optimizer,
            wrapped_optimizer=optimizer,
            train_dataloader=unwrapped_train_dataloader,
            wrapped_train_dataloader=train_dataloader,
            val_dataloader=unwrapped_val_dataloader,
            wrapped_val_dataloader=val_dataloader,
            scheduler=unwrapped_scheduler,
            wrapped_scheduler=scheduler,
        )
        self.module.device = self.accelerator.device

        self.callback.module = module
        self.callback.trainer = self
        self.callback.state = self.module.state

        self.callback.on_fit_start()

        if (
            self.eval_when_start
            and "evaluations_done" in status_dict
            and status_dict["evaluations_done"] == 0
            and DEBUG_MODE < 5
        ):
            self._eval(module, model, val_dataloader, status_dict, 0, self.hps.epochs, disable_train_loss=True)

        NUM_TRAIN_STEPS = len(self.train_dataloader)
        cleanup()
        first_epoch = True
        epoch = None
        try:
            for epoch in range(status_dict["epoch"], self.hps.epochs):
                status_dict["epoch"] = epoch
                self.monitor.log_epoch()
                initial_step = 0
                set_seed(epoch)
                if self.shuffle_train:
                    train_dataloader.set_epoch(epoch)
                if first_epoch and "skip_batches" in status_dict:
                    _train_dataloader = self.accelerator.skip_first_batches(
                        train_dataloader, status_dict["skip_batches"]
                    )
                    initial_step = status_dict["skip_batches"]
                else:
                    _train_dataloader = train_dataloader
                cleanup()
                model.train()
                self.callback.on_epoch_start()
                for step, batch in tqdm(
                    iterable=enumerate(_train_dataloader, initial_step),
                    total=len(train_dataloader),
                    initial=initial_step,
                    desc=f"Epoch {epoch}/{self.hps.epochs}",
                    unit="batch",
                ):
                    status_dict["epoch_step"] = step
                    CHECKPOINT_EVERY_N_STEPS = (
                        _CHECKPOINT_EVERY_N_STEPS and (status_dict["global_step"] + 1) % self.checkpoint_every == 0
                    )
                    EVALUATION_EVERY_N_STEPS = (
                        _EVALUATION_EVERY_N_STEPS
                        and (status_dict["global_step"] + 1) % self.evaluate_every_n_steps == 0
                    )
                    if status_dict["global_step"] % self.log_every == 0:
                        status_dict["learning_rate"] = (
                            scheduler.get_last_lr()[-1] if scheduler is not None else optimizer.param_groups[0]["lr"]
                        )
                        self.monitor.log_learning_rate()
                        self.monitor.log_cpu_utilization()
                        self.monitor.log_gpu_utilization()

                    self._train_logic(module, optimizer, batch, scheduler, status_dict)

                    if (
                        self.cleanup_cache_every_n_steps is not None
                        and status_dict["global_step"] % self.cleanup_cache_every_n_steps == 0
                    ):
                        cleanup()

                    if CHECKPOINT_EVERY_N_STEPS and DEBUG_MODE < 3:
                        self._save_checkpoint(
                            epoch, status_dict["epoch_step"] + 1, status_dict, status_dict["epoch_step"] + 1
                        )

                    if EVALUATION_EVERY_N_STEPS and DEBUG_MODE < 5:
                        self._eval(module, model, val_dataloader, status_dict, epoch, self.hps.epochs)
                        CHECKPOINT_AFTER_EVALUATION = (
                            _CHECKPOINT_AFTER_EVALUATION
                            and (status_dict["evaluations_done"] + 1) % self.checkpoint_every == 0
                        )
                        if CHECKPOINT_AFTER_EVALUATION and DEBUG_MODE < 3:
                            self._save_checkpoint(
                                epoch, status_dict["epoch_step"] + 1, status_dict, status_dict["epoch_step"] + 1
                            )
                    self.accelerator.wait_for_everyone()

                CHECKPOINT_WHEN_EPOCH_ENDS = (
                    _CHECKPOINT_WHEN_EPOCH_ENDS and (epoch + 1) % self.checkpoint_every == 0
                ) or (
                    _CHECKPOINT_AFTER_EVALUATION and (status_dict["evaluations_done"] + 1) % self.checkpoint_every == 0
                )

                if self.report_train_loss_per_epoch:
                    loss_report = self.train_track_loss / NUM_TRAIN_STEPS
                    loss_report = self.accelerator.reduce(loss_report, reduction="mean") * self.grad_accumulation_steps
                    status_dict["train_loss"] = loss_report.item()
                    self.monitor.log_train_loss(epoch=status_dict["epoch"])
                    # reset track loss
                    self.train_track_loss.zero_()

                if self.evaluate_every_n_steps is None and DEBUG_MODE < 5:
                    self._eval(module, model, val_dataloader, status_dict, epoch, self.hps.epochs)

                if CHECKPOINT_WHEN_EPOCH_ENDS and DEBUG_MODE < 3:
                    self._save_checkpoint(epoch + 1, 0, status_dict, 0)

                first_epoch = False
                cleanup()

            if (
                self.eval_when_finish
                and self.evaluate_every_n_steps is not None
                and epoch is not None
                and DEBUG_MODE < 5
            ):
                self._eval(module, model, val_dataloader, status_dict, epoch, self.hps.epochs, disable_train_loss=True)

            self.callback.on_epoch_end()
        except RuntimeError as e:
            self.callback.on_runtime_error(e)
            exception_str = str(e).lower()
            if "out of memory" in exception_str:
                self.callback.on_cuda_out_of_memory(e)

            if (
                "out of memory" in exception_str
                and any(handler in self.handlers for handler in [Handler.CUDA_OUT_OF_MEMORY, Handler.ALL])
                and DEBUG_MODE < 3
            ):
                self.accelerator.print(time_prefix(), "Forcing checkpointing due to CudaOutOfMemory error.")
                self._save_checkpoint(epoch, status_dict["epoch_step"], status_dict, status_dict["epoch_step"])
            elif any(handler in self.handlers for handler in [Handler.ANY, Handler.ALL]) and DEBUG_MODE < 3:
                self.accelerator.print(time_prefix(), "Forcing checkpointing due to a RunTime error.")
                self._save_checkpoint(epoch, status_dict["epoch_step"], status_dict, status_dict["epoch_step"])
            else:
                self.accelerator.print(e)
                traceback.print_exc()
        except KeyboardInterrupt as e:
            self.callback.on_runtime_error(e)
            if any(handler in self.handlers for handler in [Handler.KEYBOARD, Handler.ALL]) and DEBUG_MODE < 3:
                self.accelerator.print(time_prefix(), "Forcing checkpointing due to manual keyboard interrupt.")
                self._save_checkpoint(epoch, status_dict["epoch_step"], status_dict, status_dict["epoch_step"])
            else:
                self.accelerator.print(time_prefix(), "Manual keyboard interrupt.")
                traceback.print_exc()
        except Exception as e:
            self.callback.on_exception(e)
            if any(handler in self.handlers for handler in [Handler.ANY, Handler.ALL]) and DEBUG_MODE < 3:
                self.accelerator.print(time_prefix(), "Forcing checkpointing due to an exception.")
                self._save_checkpoint(epoch, status_dict["epoch_step"], status_dict, status_dict["epoch_step"])
            else:
                self.accelerator.print(e)
                traceback.print_exc()

        if epoch is None:
            raise RuntimeError(
                "Apparently you are trying to resume a training process that has already been finished."
            )
        self.callback.on_fit_end()
        self.accelerator.end_training()

    @torch.inference_mode()
    def _eval(
        self,
        module,
        model,
        val_dataloader,
        status_dict,
        epoch,
        epochs,
        disable_train_loss=False,
        disable_val_loss=False,
    ):
        cleanup()
        model.eval()

        if val_dataloader is not None:
            self.callback.on_evaluation_start()
            set_seed(epoch)
            if self.shuffle_validation:
                val_dataloader.set_epoch(epoch)
            for batch in tqdm(
                iterable=val_dataloader,
                total=len(val_dataloader),
                desc=f"Evaluating in Epoch {epoch}/{epochs}",
                unit="batch",
            ):
                self._validation_logic(module, batch, status_dict)

            # only rank 0 will be in charge of calculating metrics to avoid system overhead
            if self.accelerator.is_main_process:
                for metric in self.metrics:
                    metric_dict = metric._compute()

                    for m, v in metric_dict.items():
                        if not isinstance(v, (float, int)):
                            continue
                        status_dict["additional_metrics"][m] = v

            self._log_val_loss(status_dict, total=len(val_dataloader))
            self.monitor.log_additional_metrics()
            status_dict["evaluations_done"] += 1

        self.callback.on_evaluation_end()

        model.train()
        cleanup()

        if self.model_saving is not None and DEBUG_MODE < 3:
            self._save_model_on_criteria(model, status_dict, disable_train_loss, disable_val_loss)

    def _train_logic(self, module, optimizer, batch, scheduler, status_dict):
        self.callback.on_before_training_step(batch)
        loss = module.training_step(batch)
        self.callback.on_after_training_step()
        should_step = (status_dict["epoch_step"] + 1) % self.grad_accumulation_steps == 0
        loss = loss / self.grad_accumulation_steps

        with torch.inference_mode():
            detached_loss = loss.detach()
            self.train_total_loss += detached_loss
            self.train_track_loss += detached_loss
            status_dict["train_track_loss"] = self.train_track_loss

        log_every = self.log_every * self.grad_accumulation_steps
        if not self.report_train_loss_per_epoch and status_dict["global_step"] % log_every == 0:
            loss_report = self.train_track_loss / log_every
            loss_report = self.accelerator.reduce(loss_report, reduction="mean") * self.grad_accumulation_steps
            status_dict["train_loss"] = loss_report.item()
            self.monitor.log_train_loss()
            self.train_track_loss.zero_()  # reset track loss

        if self.accelerator.distributed_type == DistributedType.MULTI_GPU:
            self.model.require_backward_grad_sync = should_step  # for gradient sync when using DDP

        if not self.module._extended:
            self.callback.on_before_backward(loss)
            self.accelerator.backward(loss)
            self.callback.on_after_backward()

        if self.accelerator.sync_gradients and self.accelerator.is_main_process:
            norm = None
            if self.clip_grad is not None:
                norm = self.accelerator.clip_grad_norm_(self.model.parameters(), max_norm=self.clip_grad)
            elif self.monitor.grad_norm:
                norm = self._get_grad_norm()

            if norm is not None and self.monitor.grad_norm:
                status_dict["grad_norm"] = norm
                self.monitor.log_grad_norm()

        if should_step and not self.module._extended:
            self.callback.on_before_optimizer_step(optimizer)
            optimizer.step()
            self.callback.on_after_optimizer_step(optimizer)
            if scheduler is not None:
                self.callback.on_before_scheduler_step(scheduler)
                scheduler.step()
                self.callback.on_after_scheduler_step(scheduler)

            self.callback.on_before_zero_grad(optimizer)
            optimizer.zero_grad(set_to_none=self.set_to_none)
            self.callback.on_after_zero_grad(optimizer)

        status_dict["global_step"] += 1

    def _track_val_loss(self, loss):
        detached_loss = loss.detach()
        self.val_total_loss += detached_loss
        self.val_track_loss += detached_loss

    def _log_val_loss(self, status_dict, total):
        loss_report = self.val_track_loss / total
        loss_report = self.accelerator.reduce(loss_report, reduction="mean")
        status_dict["validation_loss"] = loss_report.item()
        self.monitor.log_validation_loss()
        self.val_track_loss.zero_()  # reset val loss

    def _validation_logic(self, module, batch, status_dict):
        self.callback.on_before_validation_step(batch)
        metrics_dict = module.validation_step(batch)
        self.callback.on_after_validation_step()

        self._track_val_loss(metrics_dict["loss"])

        for metric in self.metrics:
            if self.gather_none and metric.name not in metrics_dict:
                metric_compute_arguments = None, None
            else:
                metric_compute_arguments = metrics_dict[metric.name]

            if self.gather_none:
                remainder = self.accelerator.gradient_state.remainder
                metric_compute_arguments = (
                    *(
                        gather_into_single_process(arg, remainder=remainder) for arg in metric_compute_arguments
                    ),  # leave it as tuple
                )
            else:
                metric_compute_arguments = (
                    *(
                        self.accelerator.gather_for_metrics(arg) for arg in metric_compute_arguments
                    ),  # leave it as tuple
                )

            if self.accelerator.is_main_process and metric_compute_arguments[0] is not None:
                metric.add_batch(*metric_compute_arguments)

        status_dict["eval_global_step"] += 1

    def _save_model(self, model, status_dict, wait_for_everyone=True, model_path=None):
        self.accelerator.print(time_prefix(), "Saving model...")

        model_path = self.model_path if model_path is None else model_path
        os.makedirs(model_path, exist_ok=True)  # re-create folder in case it was deleted

        PATH_DOES_NOT_EXIST = not os.path.exists(self.model_path) and self.accelerator.is_main_process
        if PATH_DOES_NOT_EXIST and not wait_for_everyone:
            os.makedirs(model_path, exist_ok=True)
        if wait_for_everyone:
            if PATH_DOES_NOT_EXIST:
                os.makedirs(model_path, exist_ok=True)
            self.accelerator.wait_for_everyone()

        unwrapped_model = self.accelerator.unwrap_model(model)
        state_dict = unwrapped_model.state_dict() if not self.compile else unwrapped_model._orig_mod.state_dict()
        if hasattr(unwrapped_model, "save_pretrained"):
            unwrapped_model.save_pretrained(
                model_path,
                is_main_process=self.accelerator.is_main_process,
                state_dict=state_dict,
                max_shard_size=self.max_shard_size,
                save_function=self.accelerator.save,
                safe_serialization=self.safe_serialization,
            )
        else:
            self.accelerator.save(
                state_dict, f"{model_path}/pytorch_model.pt", safe_serialization=self.safe_serialization
            )

        if self.accelerator.is_main_process:
            save_status(status_dict, to=f"{model_path}/{STATUS_PATH}")

        self.accelerator.print(time_prefix(), "Model saved.")

    def _save_model_on_criteria(self, model, status_dict, disable_train_loss=False, disable_val_loss=False):
        # use 'disable_train_loss' when evaluating at the beginning
        if self.model_saving is None:
            return

        self.accelerator.wait_for_everyone()

        train_length = (
            self.evaluate_every_n_steps if self.evaluate_every_n_steps is not None else len(self.train_dataloader)
        )

        avg_valid_loss = self.val_total_loss / (len(self.val_dataloader) if self.val_dataloader is not None else -1)
        avg_valid_loss = self.accelerator.reduce(avg_valid_loss, reduction="mean")
        avg_train_loss = self.train_total_loss / train_length
        avg_train_loss = self.accelerator.reduce(avg_train_loss, reduction="mean")

        if self.accelerator.is_main_process:
            avg_valid_loss = avg_valid_loss.item()
            avg_train_loss = avg_train_loss.item()

            saving_criteria = {}
            for metric in self.metrics:
                best_metric_str = f"best_{metric.main_metric}"
                if best_metric_str not in status_dict:
                    status_dict[best_metric_str] = -1

                prev = status_dict[best_metric_str]
                new = status_dict["additional_metrics"][metric.main_metric]
                compare = operator_map[metric.comparator]
                is_better = compare(new, prev)
                best = new if is_better else prev

                saving_criteria[best_metric_str] = (
                    is_better and new < self.model_saving_below and new > self.model_saving_above
                )
                status_dict[best_metric_str] = best

            if not disable_val_loss:
                saving_criteria["best_valid_loss"] = (
                    avg_valid_loss < status_dict["best_valid_loss"]
                    and avg_valid_loss < self.model_saving_below
                    and avg_valid_loss > self.model_saving_above
                )
                status_dict["best_valid_loss"] = (
                    avg_valid_loss
                    if avg_valid_loss < status_dict["best_valid_loss"]
                    else status_dict["best_valid_loss"]
                )

            if not disable_train_loss:
                saving_criteria["best_train_loss"] = (
                    avg_train_loss < status_dict["best_train_loss"]
                    and avg_train_loss < self.model_saving_below
                    and avg_train_loss > self.model_saving_above
                )
                status_dict["best_train_loss"] = (
                    avg_train_loss
                    if avg_train_loss < status_dict["best_train_loss"]
                    else status_dict["best_train_loss"]
                )

            saving_criteria["always"] = True

            for model_saving in self.model_saving:
                if saving_criteria[model_saving]:
                    model_path = f"{self.model_path}/{model_saving}"
                    self._save_model(model, status_dict, wait_for_everyone=False, model_path=model_path)

        self.val_total_loss.zero_()  # reset val total loss
        self.train_total_loss.zero_()  # reset train total loss

    def _save_checkpoint(self, epoch, epoch_step, status_dict, skip_batches):
        self.callback.on_save_checkpoint()
        self.accelerator.print(time_prefix(), "Saving checkpoint...")
        if self.accelerator.is_main_process:
            if not os.path.exists(self.checkpoint):
                os.makedirs(self.checkpoint, exist_ok=True)
            if not os.path.exists(f"{self.checkpoint}/{CHECKPOINT_PATH}"):
                os.makedirs(f"{self.checkpoint}/{CHECKPOINT_PATH}", exist_ok=True)
        self.accelerator.wait_for_everyone()
        self.accelerator.save_state(f"{self.checkpoint}/{CHECKPOINT_PATH}", safe_serialization=self.safe_serialization)
        if self.accelerator.is_main_process:
            status = status_dict.copy()
            status["epoch"] = epoch
            status["epoch_step"] = epoch_step
            if self.checkpoint_strat == "step" or (
                self.checkpoint_strat == "eval" and self.evaluate_every_n_steps is not None
            ):
                status["skip_batches"] = skip_batches
            save_status(status, to=f"{self.checkpoint}/{STATUS_PATH}")
        self.accelerator.print(time_prefix(), "Checkpoint saved.")

    def _get_optimizer(self, model):
        optimizer = self.hps.optim
        fused_available = "fused" in inspect.signature(optimizer).parameters
        optim_kwargs = self.hps.optim_kwargs
        optim_kwargs["fused"] = fused_available and "cuda" in self.accelerator.device.type

        filtered_kwargs = self._filter_kwargs(optim_kwargs, optimizer)

        return optimizer(model.parameters(), **filtered_kwargs)

    def _filter_kwargs(self, _kwargs: dict, fn):
        try:
            return {k: v for k, v in _kwargs.items() if k in fn.__init__.__code__.co_varnames}
        except AttributeError:
            signature = inspect.signature(fn)
            parameters = list(signature.parameters.keys())
            return {k: v for k, v in _kwargs.items() if k in parameters}

    def _get_scheduler(self, optimizer, last_epoch, steps_per_epoch, epochs):
        steps_per_epoch = round(steps_per_epoch / self.grad_accumulation_steps)
        schlr_kwargs = self.hps.scheduler_kwargs
        schlr_kwargs["last_epoch"] = last_epoch
        schlr_kwargs["steps_per_epoch"] = steps_per_epoch
        total_steps = steps_per_epoch * epochs
        schlr_kwargs["num_training_steps"] = total_steps
        schlr_kwargs["epochs"] = epochs
        if "num_warmup_steps" in schlr_kwargs and isinstance(schlr_kwargs["num_warmup_steps"], float):
            if schlr_kwargs["num_warmup_steps"] < 0.0 or schlr_kwargs["num_warmup_steps"] > 1.0:
                raise ValueError(
                    "If 'num_warmup_steps' is a ratio (float value), it needs to be a value between 0 and 1."
                )
            schlr_kwargs["num_warmup_steps"] = round(total_steps * schlr_kwargs["num_warmup_steps"])
        elif "warmup_ratio" in schlr_kwargs:
            if schlr_kwargs["warmup_ratio"] > 1.0:
                raise ValueError(
                    "'warmup_ratio' value in scheduler configuration needs to be a value between 0 and 1."
                )
            schlr_kwargs["num_warmup_steps"] = round(total_steps * schlr_kwargs["warmup_ratio"])

        scheduler = self.hps.scheduler
        filtered_kwargs = self._filter_kwargs(schlr_kwargs, scheduler)

        return scheduler(optimizer, **filtered_kwargs)

    def _initialize_trackers(self):
        if self.accelerator.is_main_process:
            for logger in self.log_with:
                if logger.tracker == LoggerType.MLFLOW and is_url(self.logging_dir):
                    import mlflow

                    mlflow.set_tracking_uri(self.logging_dir)
                    break

    def _get_collate_fn_pipeline(self, new_collate_fn):
        def collate_fns(batch):
            for collate_fn in new_collate_fn:
                batch = collate_fn(batch)

            return batch

        return collate_fns

    def _get_grad_norm(self) -> float:
        return math.sqrt(
            sum([torch.norm(p.grad.detach()) ** 2 for p in self.model.parameters() if p.grad is not None]).item()
        )



# Copyright 2025 ghanvert. All rights reserved.
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#     http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.

from abc import ABC

import torch
from torch.optim import Optimizer
from torch.optim.lr_scheduler import LRScheduler
from typing_extensions import Any, override

from .modules import AcceleratorModule
from .states import TrainingState


class Callback(ABC):
    """
    Callback module containing different callback functions for different
    stages of the traininig process.

    NOTE: Every callback function will run on every process. If you want your
    callback functions to only run on a single process, make sure to import
    `accmt.decorators` for different function decorators.

    Attributes:
        module (`AcceleratorModule`):
            Training module.
        trainer (`Trainer`):
            Defined `Trainer` class.
        state (`TrainingState`):
            Module's `TrainingState` class.

    Methods:
        on_fit_start (*optional*):
            Callback when training process starts.
        on_fit_end (*optional*):
            Callback when training process ends.
        on_before_backward (*optional*):
            Callback before engine's backward.
        on_after_backward (*optional*):
            Callback after engine's backward.
        on_before_optimizer_step (*optional*):
            Callback before optimizers steps.
        on_after_optimizer_step (*optional*):
            Callback after optimizer steps.
        on_before_scheduler_step (*optional*):
            Callback before scheduler steps:
        on_after_scheduler_step (*optional*):
            Callback after scheduler steps.
        on_before_zero_grad (*optional*):
            Callback before optimizer resets gradients.
        on_after_zero_grad (*optional*):
            Callback after optimizer resets gradients.
        on_runtime_error (*optional*):
            Callback when process raises a `RunTimeError` exception.
        on_cuda_out_of_memory (*optional*):
            Callback when process raises a `RunTimeError` exception with
            CUDA Out Of Memory.
        on_keyboard_interrupt (*optional*):
            Callback when process raises a `KeyboardInterrupt` exception.
        on_exception (*optional*):
            Callback when process raises any other `Exception` different than
            `RuntimeError` and `KeyboardInterrupt`
        on_resume (*optional*):
            Callback when resuming training process.
        on_save_checkpoint (*optional*):
            Callback when saving checkpoint.
        on_before_training_step (*optional*):
            Callback before `training_step` function.
        on_after_training_step (*optional*):
            Callback after `training_step` function.
        on_before_validation_step (*optional*):
            Callback before `validation_step` function.
        on_after_validation_step (*optional*):
            Callback after `validation_step` function.
        on_epoch_start (*optional*):
            Callback when an epoch starts.
        on_epoch_end (*optional*):
            Callback when an epoch ends.
        on_evaluation_start (*optional*):
            Callback when evaluation starts.
        on_evaluation_end (*optional*):
            Callback when evaluation ends.

    """

    module: AcceleratorModule = None
    trainer = None
    state: TrainingState = None

    @override
    def on_fit_start(self):
        """Callback when training process starts."""

    @override
    def on_fit_end(self):
        """Callback when training process ends."""

    @override
    def on_before_backward(self, loss: torch.Tensor):
        """
        Callback before engine's backward.

        Args:
            loss (`torch.Tensor`):
                Scalar loss tensor.
        """

    @override
    def on_after_backward(self):
        """Callback after engine's backward."""

    @override
    def on_before_optimizer_step(self, optimizer: Optimizer):
        """
        Callback before optimizers steps.

        Args:
            optimizer (`Optimizer`):
                Wrapped optimizer.
        """

    @override
    def on_after_optimizer_step(self, optimizer: Optimizer):
        """
        Callback after optimizer steps.

        Args:
            optimizer (`Optimizer`):
                Wrapped optimizer.
        """

    @override
    def on_before_scheduler_step(self, scheduler: LRScheduler):
        """
        Callback before scheduler steps:

        Args:
            scheduler (`LRScheduler`):
                Wrapped scheduler.
        """

    @override
    def on_after_scheduler_step(self, scheduler: LRScheduler):
        """
        Callback after scheduler steps.

        Args:
            scheduler (`LRScheduler`):
                Wrapped scheduler.
        """

    @override
    def on_before_zero_grad(self, optimizer: Optimizer):
        """
        Callback before optimizer resets gradients.

        Args:
            optimizer (`Optimizer`):
                Wrapped optimizer.
        """

    @override
    def on_after_zero_grad(self, optimizer: Optimizer):
        """
        Callback after optimizer resets gradients.

        Args:
            optimizer (`Optimizer`):
                Wrapped optimizer.
        """

    @override
    def on_runtime_error(self, exception: Exception):
        """
        Callback when process raises a `RunTimeError` exception.

        Args:
            exception (`Exception`):
                Raised exception.
        """

    @override
    def on_cuda_out_of_memory(self, exception: Exception):
        """
        Callback when process raises a `RunTimeError` exception with
        CUDA Out Of Memory.

        Args:
            exception (`Exception`):
                Raised exception.
        """

    @override
    def on_keyboard_interrupt(self, exception: Exception):
        """
        Callback when process raises a `KeyboardInterrupt` exception.

        Args:
            exception (`Exception`):
                Raised exception.
        """
        pass

    @override
    def on_exception(self, exception: Exception):
        """
        Callback when process raises any other `Exception` different than
        `RuntimeError` and `KeyboardInterrupt`

        Args:
            exception (`Exception`):
                Raised exception.
        """

    @override
    def on_resume(self):
        """Callback when resuming training process."""

    @override
    def on_save_checkpoint(self):
        """Callback when saving checkpoint."""

    @override
    def on_before_training_step(self, batch: Any):
        """
        Callback before `training_step` function.

        Args:
            batch (`Any`):
                Dataloader's batch.
        """

    @override
    def on_after_training_step(self):
        """Callback after `training_step` function."""

    @override
    def on_before_validation_step(self, batch: Any):
        """
        Callback before `validation_step` function.

        Args:
            batch (`Any`):
                Dataloader's batch.
        """

    @override
    def on_after_validation_step(self):
        """Callback after `validation_step` function."""

    @override
    def on_epoch_start(self):
        """Callback when an epoch starts."""

    @override
    def on_epoch_end(self):
        """Callback when an epoch ends."""

    @override
    def on_evaluation_start(self):
        """Callback when evaluation starts."""

    @override
    def on_evaluation_end(self):
        """Callback when evaluation ends."""



# Copyright 2025 ghanvert. All rights reserved.
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#     http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.

from typing import Any, Optional, Union

import torch.nn as nn
from torch.optim import Optimizer
from torch.optim.lr_scheduler import LRScheduler
from torch.utils.data import DataLoader


class TrainingState:
    def __init__(
        self,
        model: Union[nn.Module, Any],
        wrapped_model: Union[nn.Module, Any],
        teacher: Union[nn.Module, Any],
        wrapped_teacher: Union[nn.Module, Any],
        optimizer: Optimizer,
        wrapped_optimizer: Union[Optimizer, Any],
        train_dataloader: DataLoader,
        wrapped_train_dataloader: Union[DataLoader, Any],
        val_dataloader: Optional[DataLoader] = None,
        wrapped_val_dataloader: Optional[Union[DataLoader, Any]] = None,
        scheduler: Optional[LRScheduler] = None,
        wrapped_scheduler: Optional[Union[LRScheduler, Any]] = None,
    ):
        self.model = model
        self.wrapped_model = wrapped_model
        self.teacher = teacher
        self.wrapped_teacher = wrapped_teacher
        self.optimizer = optimizer
        self.wrapped_optimizer = wrapped_optimizer
        self.train_dataloader = train_dataloader
        self.wrapped_train_dataloader = wrapped_train_dataloader
        self.val_dataloader = val_dataloader
        self.wrapped_val_dataloader = wrapped_val_dataloader
        self.scheduler = scheduler
        self.wrapped_scheduler = wrapped_scheduler



# Copyright 2025 ghanvert. All rights reserved.
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#     http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.

import datetime
import gc
import json
import operator
import os
import re
import sys
import warnings
from contextlib import contextmanager

import pandas as pd
import torch


units = {
    "epoch": {"epoch", "ep", "epochs", "eps"},
    "step": {"step", "st", "steps", "sts"},
    "eval": {"evaluation", "eval", "evaluations", "evals"},
}


def is_url(string):
    if string in ["localhost", "127.0.0.1"]:
        return True

    url_regex = re.compile(
        r"^(?:http|ftp)s?://"
        r"(?:(?:[A-Z0-9](?:[A-Z0-9-]{0,61}[A-Z0-9])?\.)+(?:[A-Z]{2,6}\.?|[A-Z0-9-]{2,}\.?)|"
        r"localhost|"
        r"\d{1,3}\.\d{1,3}\.\d{1,3}\.\d{1,3})"
        r"(?::\d+)?"
        r"(?:/?|[/?]\S+)$",
        re.IGNORECASE,
    )
    return re.match(url_regex, string) is not None


def get_number_and_unit(string: str):
    match = re.match(r"(\d+)(\D+)", string)

    if match:
        number = int(match.group(1))
        text = match.group(2).strip().lower()
    else:
        number = 1
        text = string.strip().lower()

    unit = None
    for k, v in units.items():
        if text in v:
            unit = k
            break

    return number, unit


def combine_dicts(*dicts):
    combined = {}
    for d in dicts:
        combined.update(d)
    return combined


def divide_list(lst: list, parts: int):
    k, m = divmod(len(lst), parts)
    return [lst[i * k + min(i, m) : (i + 1) * k + min(i + 1, m)] for i in range(parts)]


def time_prefix():
    return "[" + datetime.datetime.now().strftime("%Y-%m-%d %H:%M:%S,%f")[:-3] + "]"


PANDAS_READER_MAP = {
    "csv": pd.read_csv,
    "xlsx": pd.read_excel,
    "xml": pd.read_xml,
    "feather": pd.read_feather,
    "parquet": pd.read_parquet,
    "pickle": pd.read_pickle,
    "pkl": pd.read_pickle,
}


def save_status(status: dict, to: str):
    for key in status.keys():
        if isinstance(status[key], torch.Tensor):
            status[key] = status[key].item()
    json_string = json.dumps(status, indent=4)
    open(to, "w").write(json_string)


def read_status(path: str) -> dict:
    return json.load(open(path))


@contextmanager
def suppress_print_and_warnings(verbose=False):
    if not verbose:
        original_stdout = sys.stdout
        sys.stdout = open(os.devnull, "w")
        with warnings.catch_warnings():
            warnings.simplefilter("ignore")
            try:
                yield
            finally:
                sys.stdout.close()
                sys.stdout = original_stdout
    else:
        yield


def cleanup():
    gc.collect()
    torch.cuda.empty_cache()


operator_map = {"<": operator.lt, "<=": operator.le, ">": operator.gt, ">=": operator.ge, "==": operator.eq}



# Copyright 2025 ghanvert. All rights reserved.
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#     http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.

import os
from dataclasses import dataclass
from typing import Optional, Union

import psutil
import torch
from accelerate import Accelerator, DistributedType


@dataclass
class Monitor:
    """
    Class to set metrics to monitor during training using a tracker (if implemented).

    Args:
        learning_rate (`bool`, *optional*, defaults to `False`):
            Monitor learning rate.
        epoch (`bool`, *optional*, defaults to `True`):
            Monitor current epoch.
        train_loss (`bool`, *optional*, defaults to `True`):
            Monitor training loss.
        validation_loss (`bool`, *optional*, defaults to `True`):
            Monitor validation loss.
        accuracy (`bool`, *optional*, defaults to `True`):
            Monitor accuracy if implemented.
        grad_norm (`bool`, *optional*, defaults to `False`):
            This will enable monitoring for gradient normalization. This feature is not yet supported
            when running with DeepSpeed.
        gpu_utilization (`bool`, *optional*, defaults to `False`):
            Monitor GPU utilization in GB. It only reports GPU from main process (for now).
        cpu_utilization (`bool`, *optional*, defaults to `False`):
            Monitor CPU utilization in GB. It only reports CPU from main process (for now)
        val_equal_train (`bool`, *optional*, defaults to `True`):
            When reporting validation loss and accuracy, its step will be equal to train loss. If set to
            `False`, validation step will be equal to the number of evaluations done (starting at 0).
            This argument is only valid when `report_loss_after_eval` is set to `True`.
    """

    def __init__(
        self,
        learning_rate: bool = False,
        epoch: bool = True,
        train_loss: bool = True,
        validation_loss: bool = True,
        additional_metrics: bool = True,
        grad_norm: bool = False,
        gpu_utilization: bool = False,
        cpu_utilization: bool = False,
        val_equal_train: bool = True,
    ):
        self.learning_rate = learning_rate
        self.epoch = epoch
        self.train_loss = train_loss
        self.validation_loss = validation_loss
        self.additional_metrics = additional_metrics
        self.grad_norm = grad_norm
        self.gpu_utilization = gpu_utilization
        self.cpu_utilization = cpu_utilization
        self.val_equal_train = val_equal_train
        self.status_dict = None
        self.accelerator = None
        self.train_loss_name = None
        self.validation_loss_name = None
        self._do_tracking = True
        self._debug_mode = 0

    @classmethod
    def from_config(cls, config: Union[str, dict]):
        """
        Load a monitor configuration from a file or a dictionary.

        Args:
            config (`str` or `dict`):
                Path to a file or dictionary containing kwargs for Monitor constructor. The file can
                be YAML or JSON.
        """
        assert config is None or isinstance(config, (str, dict)), f"{config} is not of type 'str' or 'dict'."
        if isinstance(config, str):
            import yaml

            config = yaml.safe_load(open(config))
        elif config is None:
            config = {}

        return Monitor(**config)

    def _set_extra(self, accelerator: Accelerator, status_dict: dict, train_loss_name: str, validation_loss_name: str):
        self.accelerator = accelerator
        self.status_dict = status_dict
        self.train_loss_name = train_loss_name
        self.validation_loss_name = validation_loss_name

    def log_learning_rate(self):
        if self.learning_rate and self.accelerator.is_main_process and self._do_tracking and self._debug_mode < 1:
            self.accelerator.log(
                {"learning_rate": self.status_dict["learning_rate"]}, step=self.status_dict["global_step"]
            )

    def log_epoch(self):
        if self.epoch and self.accelerator.is_main_process and self._do_tracking and self._debug_mode >= 1:
            self.accelerator.log({"epoch": self.status_dict["epoch"]}, step=self.status_dict["epoch"])

    def log_train_loss(self, epoch: Optional[int] = None):
        if self.train_loss and self.accelerator.is_main_process and self._do_tracking and self._debug_mode < 1:
            loss = self.status_dict["train_loss"]
            loss = loss.item() if isinstance(loss, torch.Tensor) else loss
            step = self.status_dict["global_step"] if epoch is None else epoch
            self.accelerator.log({self.train_loss_name: loss}, step=step)

    def log_validation_loss(self):
        if self.validation_loss and self.accelerator.is_main_process and self._do_tracking and self._debug_mode < 1:
            step = (
                self.status_dict["eval_global_step"] if self.val_equal_train else self.status_dict["evaluations_done"]
            )
            loss = self.status_dict["validation_loss"]
            loss = loss.item() if isinstance(loss, torch.Tensor) else loss
            self.accelerator.log({self.validation_loss_name: loss}, step=step)

    def log_additional_metrics(self):
        if self.additional_metrics and self.accelerator.is_main_process and self._do_tracking and self._debug_mode < 1:
            step = (
                self.status_dict["eval_global_step"] if self.val_equal_train else self.status_dict["evaluations_done"]
            )
            for metric, value in self.status_dict["additional_metrics"].items():
                self.accelerator.log({metric: value}, step=step)

    def log_gpu_utilization(self):
        if self.gpu_utilization and self.accelerator.is_main_process and self._do_tracking and self._debug_mode < 1:
            if self.accelerator.distributed_type in {
                DistributedType.DEEPSPEED,
                DistributedType.FSDP,
                DistributedType.MULTI_GPU,
            }:
                device = torch.device("cuda")
                memory_allocated = torch.cuda.memory_allocated(device)
                memory_reserved = torch.cuda.memory_reserved(device)
                total_memory = (memory_allocated + memory_reserved) / (1024**3)

                self.accelerator.log({"GPU_0": total_memory}, step=self.status_dict["global_step"])

    def log_cpu_utilization(self):
        if self.cpu_utilization and self.accelerator.is_main_process and self._do_tracking and self._debug_mode < 1:
            process = psutil.Process(os.getpid())
            cpu_mem = process.memory_info().rss / (1024**3)
            self.accelerator.log({"CPU_PROCESS_0": cpu_mem}, step=self.status_dict["global_step"])

    def log_grad_norm(self):
        if (
            self.grad_norm
            and self.accelerator.is_main_process
            and "grad_norm" in self.status_dict
            and self._do_tracking
            and self._debug_mode < 1
        ):
            self.accelerator.log({"grad_norm": self.status_dict["grad_norm"]}, step=self.status_dict["global_step"])



# Copyright 2025 ghanvert. All rights reserved.
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#     http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.

from dataclasses import dataclass

from accelerate.utils import LoggerType


@dataclass
class TensorBoard:
    tracker = LoggerType.TENSORBOARD

    def init(**kwargs):
        return {"tensorboard": {**kwargs}}


@dataclass
class WandB:
    tracker = LoggerType.WANDB

    def init(**kwargs):
        return {"wandb": {**kwargs}}


@dataclass
class CometML:
    tracker = LoggerType.COMETML

    def init(**kwargs):
        return {"comet_ml": {**kwargs}}


@dataclass
class Aim:
    tracker = LoggerType.AIM

    def init(**kwargs):
        return {"aim": {**kwargs}}


@dataclass
class MLFlow:
    tracker = LoggerType.MLFLOW

    def init(**kwargs):
        return {"mlflow": {**kwargs}}


@dataclass
class ClearML:
    tracker = LoggerType.CLEARML

    def init(**kwargs):
        return {"clearml": {**kwargs}}


@dataclass
class DVCLive:
    tracker = LoggerType.DVCLIVE

    def init(**kwargs):
        return {"dvclive": {**kwargs}}



# Copyright 2025 ghanvert. All rights reserved.
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#     http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.

from dataclasses import dataclass


@dataclass
class Handler:
    """
    Properties:
        `KEYBOARD`: Handles `KeyboardInterrupt` exceptions (CTRL + C).
        `CUDA_OUT_OF_MEMORY`: Handles cuda out of memory errors derived from `RuntimeError` exceptions.
        `ANY`: Handles any other exception not considered in these properties.
        `ALL`: Handles all possible exceptions.
    """

    KEYBOARD = 1
    CUDA_OUT_OF_MEMORY = 2
    ANY = 3
    ALL = 4



# Copyright 2025 ghanvert. All rights reserved.
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#     http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.

import os
from datetime import timedelta

import torch
from accelerate import Accelerator, DataLoaderConfiguration, InitProcessGroupKwargs
from accelerate.utils import tqdm

from .callbacks import Callback
from .collate_fns import DataCollatorForLanguageModeling, DataCollatorForLongestSequence, DataCollatorForSeq2Seq
from .dataloader_samplers import TemperatureSampler
from .decorators import on_last_process, on_local_main_process, on_local_process, on_main_process, on_process
from .handlers import Handler
from .hyperparameters import HyperParameters, Optimizer, Scheduler
from .modules import AcceleratorModule, ExtendedAcceleratorModule
from .monitor import Monitor
from .tracker import Aim, ClearML, CometML, DVCLive, MLFlow, TensorBoard, WandB
from .trainer import Trainer, set_seed
from .utility import prepare, prepare_array, prepare_dataframe


def allow_tf32(flag=True):
    """Enable or disable the use of TensorFloat32."""
    torch.set_float32_matmul_precision("high" if flag else "highest")


allow_tf32()

IS_CPU = bool(int(os.environ.get("ACCMT_CPU", 0)))
IS_GPU = not IS_CPU

_init_kwargs = InitProcessGroupKwargs(timeout=timedelta(seconds=86400))
_dataloader_config = DataLoaderConfiguration(use_seedable_sampler=True)
accelerator = Accelerator(
    kwargs_handlers=[_init_kwargs],
    dataloader_config=_dataloader_config,
    step_scheduler_with_optimizer=False,
    cpu=IS_CPU,
)



# Copyright 2025 ghanvert. All rights reserved.
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#     http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.


def on_main_process(function):
    """Run on main process (RANK = 0)."""
    from . import accelerator

    return accelerator.on_main_process(function)


def on_last_process(function):
    """Run on last process (RANK = WORLD_SIZE - 1)."""
    from . import accelerator

    return accelerator.on_last_process(function)


def on_local_main_process(function):
    """Run on local main process (LOCAL_RANK = 0)."""
    from . import accelerator

    return accelerator.on_local_main_process(function)


def on_local_process(function, local_process_index: int):
    """Run on a specific local process (LOCAL_RANK = `local_process_index`)."""
    from . import accelerator

    return accelerator.on_local_process(function, local_process_index)


def on_process(function, process_index: int):
    """Run on a specific process (RANK = `process_index`)."""
    from . import accelerator

    return accelerator.on_process(function, process_index)
